/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#include "postgres.h"
#include <string.h>
#include <time.h>
#include "algstructs.h"
#include "queues.h"
#include "queuesfuncs.h"
#include "algutils.h"
#include "catalog/pg_namespace.h"
#include "utils/memutils.h"
#include "parser/parse_coerce.h"

/* a helper function to convert time var to pretty text time */
static char *
str_time(time_t tnow)
{
	static char buf[128];

	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S %Z", localtime(&tnow));
	return buf;
}

/* here we parse the relations parameter the user gives 
 * into a list of Oid of these tables */
Oid *
parseOIDs(text *relsStr, int *numOfRels)
{
	if (relsStr == NULL)
      ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errhint("Checks your relations input, e.g., 'rel1,rel2,rel3'.")));

	int			relsStrSize = VARSIZE(relsStr) - VARHDRSZ + 1;
	char		relsStrCp[relsStrSize];

	memcpy(relsStrCp, VARDATA(relsStr), relsStrSize - 1);
	relsStrCp[relsStrSize - 1] = (char) NULL;
	return parseOIDs_(relsStrCp, numOfRels);
}

/* we just convert text to str */
char *
getRelsCharStr(text *relsStr)
{
	if (relsStr == NULL)
        ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errhint("Checks your relations input, e.g., 'rel1,rel2,rel3'.")));

	int			relsStrSize = VARSIZE(relsStr) - VARHDRSZ + 1;
	char	   *relsStrCp = (char *) palloc(relsStrSize * sizeof(char));

	memcpy(relsStrCp, VARDATA(relsStr), relsStrSize - 1);
	relsStrCp[relsStrSize - 1] = (char) NULL;
	return relsStrCp;
}

/* here we parse the relations into a list of its Oids */
Oid *
parseOIDs_(char *relsStrCp, int *numOfRels)
{
	Oid			oids[MaxHeapAttributeNumber];
	const char	delimiters[] = ",";
	const char	tokenDelimiters[] = ".";
	char	   *token = NULL;
	char	   *relsStrSvpt = NULL;
	char	   *tokenSvpt = NULL;

	token = strtok_r(relsStrCp, delimiters, &relsStrSvpt);
    
	char	   *namespace = NULL;
	char	   *relname = NULL;
	int			i = -1;
	Oid			namespaceId = -1;

	while (token != NULL)
	{
		namespace = strtok_r(token, tokenDelimiters, &tokenSvpt);
		relname = strtok_r(NULL, tokenDelimiters, &tokenSvpt);
		if (relname == NULL)
		{
			relname = namespace;
			namespace = NULL;
		}
		if (relname == NULL)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errhint("Checks your relations input and delimiters ', .' , e.g., 'rel1,rel2,rel3'.")));
		i++;
		if (namespace != NULL)
			namespaceId = GetSysCacheOid(NAMESPACENAME,
										 CStringGetDatum(namespace), 0, 0, 0);
		else
			namespaceId = PG_PUBLIC_NAMESPACE;
		if (!OidIsValid(namespaceId))
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errhint("Checks your relations input possibly no such namespace ', .' , e.g., 'rel1,rel2,rel3'.")));
		if (!OidIsValid(oids[i] = get_relname_relid(relname, namespaceId)))
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errhint("Checks your relations input possibly no such relation ', .' , e.g., 'rel1,rel2,rel3'.")));
		token = strtok_r(NULL, delimiters, &relsStrSvpt);
	}
	*numOfRels = i + 1;
	Oid		   *relsOID = (Oid *) palloc(*numOfRels * sizeof(Oid));

	for (i = 0; i < *numOfRels; i++)
		relsOID[i] = oids[i];
	return relsOID;
}

/* Here we pass the inputted mapping parameter into a Dictionary (see .h file */
Dictionary *
parseAttributeDictionary(char *mappingStrCp, Oid *relsOid, int numOfRels)
{
	int			i,
				j;
	int			numAtts[numOfRels];
	int			maxNumAtts = 0;
	char	   *attName;
	int			numPotentialAttNames = 0;
	Mapping  ***relsAttsMappings = (Mapping ***) palloc(numOfRels * sizeof(Mapping **));
	int		   *numMappings = (int *) palloc(numOfRels * sizeof(int));

	for (i = 0; i < numOfRels; i++)
	{
		numAtts[i] = 0;
		numMappings[i] = 0;
		HeapTuple	attTuple;

		for (j = 1;; j++)
		{
			attName = (char *) get_attname(relsOid[i], j);
			if (attName != NULL)
			{
				attTuple = SearchSysCacheAttName(relsOid[i], attName);
				if (attTuple != NULL)
				{
					ReleaseSysCache(attTuple);
					numAtts[i]++;
				}
			}
			else
				break;
		}
		maxNumAtts += numAtts[i];
		relsAttsMappings[i] = (Mapping **) palloc(numAtts[i] * sizeof(Mapping *));
	}
    
	char	   *resultTuplePotentialAttNames[maxNumAtts];
	const char	delimiters[] = ",";
	const char	tokenDelimiters[] = "=.";
	char	   *token = NULL;
	char	   *mappingStrSvpt = NULL;
	char	   *tokenSvpt = NULL;

	token = strtok_r(mappingStrCp, delimiters, &mappingStrSvpt);
	char	   *namespace = NULL;
	char	   *relname = NULL;
	char	   *resultTupleAttName;
	char	   *relAttName;

	i = -1;
	Oid			namespaceId = -1;
	Oid			relOid = -1;
	int			relid;

	while (token != NULL)
	{
		resultTupleAttName = strtok_r(token, tokenDelimiters, &tokenSvpt);
		if (resultTupleAttName == NULL)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("problem parsing mapping string!"),
			  errhint("Check your format. e.g 'lastname=public.rel1.username'")));
		namespace = strtok_r(NULL, tokenDelimiters, &tokenSvpt);
		if (namespace == NULL)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("problem parsing mapping string!"),
			  errhint("Check your format. e.g 'lastname=public.rel1.username'")));
		relname = strtok_r(NULL, tokenDelimiters, &tokenSvpt);
		if (relname == NULL)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("problem parsing mapping string!"),
			  errhint("Check your format. e.g 'lastname=public.rel1.username'")));
		relAttName = strtok_r(NULL, tokenDelimiters, &tokenSvpt);
		if (relAttName == NULL)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("problem parsing mapping string!"),
			  errhint("Check your format. e.g 'lastname=public.rel1.username'")));
		if (namespace != NULL)
			namespaceId = GetSysCacheOid(NAMESPACENAME,
										 CStringGetDatum(namespace), 0, 0, 0);
		if (!OidIsValid(namespaceId))
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errdetail("Checks your relations input possibly no such namespace: %s",namespace)));
		if (!OidIsValid(relOid = get_relname_relid(relname, namespaceId)))
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errdetail("Checks your relations input possibly no such relation: %s",relname)));

		HeapTuple	attTuple = SearchSysCacheAttName(relOid, relAttName);

		if (attTuple != NULL)
			ReleaseSysCache(attTuple);
		else
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errdetail("Checks your relations input possibly no such relation attribute: %s",relAttName)));
	
        relid = -1;
		for (i = 0; i < numOfRels; i++)
		{
			if (relOid == relsOid[i])
			{
				relid = i;
				break;
			}
		}
		if (relid == -1)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Something is wrong with the relations inputed"),
			  errdetail("Checks your relations input possibly no such relation: %s.%s",namespace,relname)));

		int			nameLocation = -1;

		for (i = 0; i < numPotentialAttNames; i++)
			if (strcmp(resultTuplePotentialAttNames[i], resultTupleAttName) == 0)
			{
				nameLocation = i;
				break;
			}
		if (nameLocation == -1)
		{
			resultTuplePotentialAttNames[numPotentialAttNames] = resultTupleAttName;
			nameLocation = numPotentialAttNames;
			numPotentialAttNames++;
		}
        
		Mapping    *mapping = palloc(sizeof(Mapping));

		mapping->resultTuplePotnetialAttNamesID = nameLocation;
		mapping->relAttName = relAttName;
		relsAttsMappings[relid][numMappings[relid]] = mapping;
		numMappings[relid]++;
		token = strtok_r(NULL, delimiters, &mappingStrSvpt);
	}
	Dictionary *dictionary = palloc(sizeof(Dictionary));

	dictionary->resultTuplePotentialAttNames = palloc(numPotentialAttNames * sizeof(char *));
    
	for (i = 0; i < numPotentialAttNames; i++)
		dictionary->resultTuplePotentialAttNames[i] = resultTuplePotentialAttNames[i];
    
	dictionary->numPotentialAttNames = numPotentialAttNames;
	dictionary->relsAttsMappings = relsAttsMappings;
	dictionary->numMappings = numMappings;
	
    int			k;
	for (i = 0; i < numOfRels; i++)
		for (j = 0; j < numMappings[i]; j++)
		{
			for (k = 0; k < numMappings[i]; k++)
				if ((j != k) &&
					(strcmp(resultTuplePotentialAttNames[relsAttsMappings[i][j]->resultTuplePotnetialAttNamesID],
							relsAttsMappings[i][k]->relAttName) == 0))
                    ereport(ERROR,
                     (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                      errmsg("Circular mappings!"),
                      errdetail("Check your mappings. Specifically: %s, %s",
                         resultTuplePotentialAttNames[relsAttsMappings[i][j]->resultTuplePotnetialAttNamesID],
						 relsAttsMappings[i][k]->relAttName)));
			if (strcmp(resultTuplePotentialAttNames[relsAttsMappings[i][j]->resultTuplePotnetialAttNamesID],
					   relsAttsMappings[i][j]->relAttName) != 0)
			{
				HeapTuple	attTuple = SearchSysCacheAttName(relsOid[i],
															 resultTuplePotentialAttNames[relsAttsMappings[i][j]->resultTuplePotnetialAttNamesID]);

				if (attTuple != NULL)
				{
					ReleaseSysCache(attTuple);
                    ereport(ERROR,
                     (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                      errmsg("Circular mappings!"),
                      errdetail("Check your mappings. Specifically: %s",
                        resultTuplePotentialAttNames[relsAttsMappings[i][j]->resultTuplePotnetialAttNamesID])));
				}
			}
		}
	return dictionary;
}

/* we take a dictionary and map an attribute name from a relation 
 * to another name using the dictionary */
char *
mapAttName(Dictionary * dictionary, int relid, char *attname)
{
	if (dictionary == NULL)
		return attname;
	else
	{
		int			i;
		int			numMappings = dictionary->numMappings[relid];
		Mapping    *mapping = NULL;

		if (numMappings == 0)
			return attname;
		for (i = 0; i < numMappings; i++)
		{
			mapping = dictionary->relsAttsMappings[relid][i];
			if (strcmp(mapping->relAttName, attname) == 0)
			{
				return dictionary->resultTuplePotentialAttNames[mapping->resultTuplePotnetialAttNamesID];
			}
		}
		return attname;
	}
}

/* just a convinience call */
void
initialize_general_structures(PG_FUNCTION_ARGS, alg_fctx * fctx, FuncCallContext *funcctx)
{
	initialize_general_structures_(fcinfo, fctx, true, funcctx);
}

/* initialized the fd general required structures */
void
initialize_general_structures_(PG_FUNCTION_ARGS, alg_fctx * fctx, bool parsing, FuncCallContext *funcctx)
{
	unsigned int tmpTime = (unsigned) time(NULL);

	srand(tmpTime);

#ifdef DEBUG
	ereport(LOG, (errmsg(" ")));
	ereport(LOG, (errmsg("FD-DEBUG: %s", str_time(time(NULL)))));
	ereport(LOG, (errmsg("FD-DEBUG-SRAND:%d", tmpTime)));
	ereport(LOG, (errmsg("FD-DEBUG-Relations:%s", getRelsCharStr(PG_GETARG_TEXT_P(0)))));
#endif
    
	fctx->per_query_ctx = AllocSetContextCreate(CurrentMemoryContext,
												"tuplestore2",
												ALLOCSET_DEFAULT_MINSIZE,
												ALLOCSET_DEFAULT_INITSIZE,
												ALLOCSET_DEFAULT_MAXSIZE);
	fctx->nargs = PG_NARGS();
	char	   *mappingStr = NULL;

	if (fctx->nargs >= 1)
	{
		char	   *policy = NULL;

		switch (fctx->nargs)
		{
			case 9:
			case 8:
			case 7:
			case 6:
			case 5:
				{
					text	   *policyText = PG_GETARG_TEXT_P(4);
					int			policyTextSize = VARSIZE(policyText) - VARHDRSZ;

					if (policyTextSize == 0)
						policy = NULL;
					else
					{
						policy = (char *) palloc(policyTextSize * sizeof(char) + 1);
						memcpy(policy, VARDATA(policyText), policyTextSize);
						policy[policyTextSize] = '\0';
					}
				}
			case 4:
				fctx->noIndexScans = PG_GETARG_BOOL(3);
			case 3:
				if (fctx->nargs <= 3)
					fctx->noIndexScans = false;
				fctx->noDuplicates = PG_GETARG_BOOL(2);
			case 2:
				{
					text	   *mappingText = PG_GETARG_TEXT_P(1);
					int			mappingTextSize = VARSIZE(mappingText) - VARHDRSZ;

					if (mappingTextSize > 0)
					{
						mappingStr = (char *) palloc(mappingTextSize * sizeof(char) + 1);
						memcpy(mappingStr, VARDATA(mappingText), mappingTextSize);
						mappingStr[mappingTextSize] = '\0';
					}
				}
			case 1:
				if (fctx->nargs <= 2)
				{
					fctx->noIndexScans = false;
					fctx->noDuplicates = true;
				}
		}
#ifdef DEBUG
		if (policy != NULL)
			ereport(LOG, (errmsg("FD-DEBUG-Policy:%s", policy)));
#endif
		fctx->query = NULL;
		fctx->querySize = 0;
		fctx->queryParams = NULL;
		fctx->queryParamsTypes = NULL;
		fctx->queryParamsSize = 0;
        
		fctx->policyWorkNode = NULL;
		fctx->policySequence = 0;
		fctx->extendingPolicy = -1;
		if ((policy != NULL) && strcmp(policy, "majority") == 0)
			fctx->extendingPolicy = POLICY_MAJORITY;
		else if ((policy != NULL) && strcmp(policy, "mincost") == 0)
			fctx->extendingPolicy = POLICY_MINCOST;
		else if ((policy != NULL) && strcmp(policy, "random") == 0)
			fctx->extendingPolicy = POLICY_RANDOM;
		else if ((policy != NULL) && strcmp(policy, "naive") == 0)
			fctx->extendingPolicy = POLICY_NAIVE;
		else
		{
			if (policy != NULL)
              ereport(ERROR,
               (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                errmsg("No such policy!"),
                errdetail("Check out the inputed policy: %s", policy)));
			fctx->extendingPolicy = POLICY_MAJORITY;
		}
		if (policy != NULL)
			pfree(policy);
	}
    
	int			i = 0,
				j = 0,
				k = 0;
	char	   *attName = NULL;

	{
		if (parsing)
			fctx->relsOid = parseOIDs(PG_GETARG_TEXT_P(0), &fctx->numRels);
		fctx->relsPages = (int *) palloc(fctx->numRels * sizeof(int));
		fctx->relsHasIndex = (bool *) palloc(fctx->numRels * sizeof(bool));
		fctx->sortedRelsByRelsPages = (int *) palloc(fctx->numRels * sizeof(int));
        
		fctx->relsMaxTupleSize = (int *) palloc(fctx->numRels * sizeof(int));
		for (i = 0; i < fctx->numRels; i++)
			fctx->relsMaxTupleSize[i] = 0;
        
		fctx->relsNAtt = (int *) palloc(fctx->numRels * sizeof(int));
		fctx->relsAttNumsAtTupleSet
			= (int **) palloc(fctx->numRels * sizeof(int *));
		fctx->relsAttTypes = (Oid **) palloc(fctx->numRels * sizeof(Oid *));
		fctx->resultTupleAttNames = NULL;
		fctx->tupleSetAttTypes = NULL;
        
		fctx->resultTupleDesc = NULL;
		fctx->tupleSetDesc = NULL;
        
		fctx->resultTupleNAtt = 0;
		fctx->tupleSetMaxSize = RESULT_TUPLE_ASSUMED_MAX_SIZE_IN_BYTES;
		fctx->assumingResultTupleMaxSize = true;
		fctx->relsValidAttNumAtRelation = (int **) palloc(fctx->numRels * sizeof(int *));
        
		fctx->bits_scheme_graph = (RBITS *) palloc(fctx->numRels * sizeof(RBITS));
        
		fctx->dynamic_extend_cap = (adaptive_extend_upperbound *) palloc(sizeof(adaptive_extend_upperbound));
		fctx->dynamic_extend_cap->previousScore = 0;
		fctx->dynamic_extend_cap->nextUpperBound = 0;
		fctx->dynamic_extend_cap->stage = 0;
		fctx->dynamic_extend_cap->maxInputedExtendTuples = 0;
        
		fctx->printedTSet = NULL;
		fctx->startedPrintingQueue = false;
		fctx->printedTupleAddress = NULL;
	}
    
    /* sort the array of relations, i.e. their oids */
	sortOidArray(fctx->relsOid, fctx->numRels);
    
    /* Lets parse the dictionary. */
	if (mappingStr == NULL)
	{
		fctx->dictionary = NULL;
	}
	else
		fctx->dictionary = parseAttributeDictionary(mappingStr, fctx->relsOid, fctx->numRels);
    
	/* get each rel number of pages. */
	{
		Form_pg_class pgcform;
		HeapTuple	ctup;
		int			tmpRelsPages[fctx->numRels];

		for (i = 0; i < fctx->numRels; i++)
		{
			ctup = SearchSysCache(RELOID,
								  ObjectIdGetDatum(fctx->relsOid[i]),
								  0, 0, 0);
			pgcform = (Form_pg_class) GETSTRUCT(ctup);
			fctx->relsHasIndex[i] = pgcform->relhasindex;
			fctx->relsPages[i] = pgcform->relpages;
			ReleaseSysCache(ctup);
			tmpRelsPages[i] = fctx->relsPages[i];
			fctx->sortedRelsByRelsPages[i] = i;
		}

		/*
		 * Here we create  a sorted array of the relations, sorted by the
		 * number of pages they contain.
		 */
		int			swapint;

		for (i = 0; i < (fctx->numRels - 1); i++)
			for (j = (i + 1); j < fctx->numRels; j++)
				if (tmpRelsPages[i] > tmpRelsPages[j])
				{
					swapint = tmpRelsPages[i];
					tmpRelsPages[i] = tmpRelsPages[j];
					tmpRelsPages[j] = swapint;
					swapint = fctx->sortedRelsByRelsPages[i];
					fctx->sortedRelsByRelsPages[i] = fctx->sortedRelsByRelsPages[j];
					fctx->sortedRelsByRelsPages[j] = swapint;
				}
	}

	/*
	 * here we get the number of attributes for each relation And we also
	 * creating the resultTuple Attribute names. we exploit the fact that
	 * attributes of relations that are first in the order comes first in the
	 * result tuple so do not change that!
	 */
	{
		int			relsValidAttNumAtRelation[MaxHeapAttributeNumber];
		bool		foundAtt = 0;
		bool		skipped = false;
		char	   *mappedAttName = NULL;

		fctx->resultTupleAttNames = (char **) palloc(MaxHeapAttributeNumber * sizeof(char *));
		for (i = 0; i < fctx->numRels; i++)
		{
			fctx->relsNAtt[i] = 0;
			for (j = 1;; j++)
			{
				attName = (char *) get_attname(fctx->relsOid[i], j);
				skipped = false;
				if (attName != NULL)
				{
					/*
					 * here we check if an attribute was dropped by using the
					 * special SearchSysCacheAttName
					 */
					HeapTuple	attTuple = SearchSysCacheAttName(fctx->relsOid[i], attName);

					if (attTuple != NULL)
						ReleaseSysCache(attTuple);
					else
					{
						pfree(attName);
						attName = NULL;
						skipped = true;
					}
				}
                /* Ok so the attribute is valid. For example, an attribute of dropped
                 * column is not valid.*/
				if (attName != NULL)
				{
					relsValidAttNumAtRelation[fctx->relsNAtt[i]] = j;
					fctx->relsNAtt[i]++;
					foundAtt = 0;
                    /* lets map the attribute of the relation using the mappings entered
                     * by the user */
					mappedAttName = mapAttName(fctx->dictionary, i, attName);
                    /* let us see if we already added this attribute to the result tuple */
					for (k = 0; k < fctx->resultTupleNAtt; k++)
					{
						if (strcmp(fctx->resultTupleAttNames[k], mappedAttName) == 0)
						{
							foundAtt = 1;
							break;
						}
					}
					if (!foundAtt)
					{
						fctx->resultTupleAttNames[fctx->resultTupleNAtt] = mappedAttName;
						fctx->resultTupleNAtt++;
					}
				}
				else if (!skipped)
					break;
			}
			fctx->relsValidAttNumAtRelation[i] = (int *) palloc(fctx->relsNAtt[i] * sizeof(int));
			for (j = 0; j < fctx->relsNAtt[i]; j++)
				fctx->relsValidAttNumAtRelation[i][j] = relsValidAttNumAtRelation[j];
		}
		fctx->tupleSetNatt = fctx->resultTupleNAtt + 1;
	}
    
	/* lets free the wasted resources. */
	{
		char	  **resultTupleAttNamesTemp
		= (char **) palloc(fctx->resultTupleNAtt * sizeof(char *));

		for (k = 0; k < fctx->resultTupleNAtt; k++)
		{
			resultTupleAttNamesTemp[k] = fctx->resultTupleAttNames[k];
		}
		pfree(fctx->resultTupleAttNames);
		fctx->resultTupleAttNames = resultTupleAttNamesTemp;
	}

	/*
	 * here we get the types of each attribute in each relation. And we also
	 * assign the location of each attribute in each relation in the result
	 * tuple, in their respective relation search table.
	 */
	{
		char	   *mappedAttName = NULL;

		for (i = 0; i < fctx->numRels; i++)
		{
			fctx->relsAttNumsAtTupleSet[i]
				= (int *) palloc(fctx->relsNAtt[i] * sizeof(int));
			fctx->relsAttTypes[i] = (Oid *) palloc(fctx->relsNAtt[i] * sizeof(Oid));
			for (j = 0; j < fctx->relsNAtt[i]; j++)
			{
				fctx->relsAttTypes[i][j] = get_atttype(fctx->relsOid[i], fctx->relsValidAttNumAtRelation[i][j]);
				attName = (char *) get_attname(fctx->relsOid[i], fctx->relsValidAttNumAtRelation[i][j]);
				mappedAttName = mapAttName(fctx->dictionary, i, attName);
				for (k = 0; k < fctx->resultTupleNAtt; k++)
				{
					if (strcmp(fctx->resultTupleAttNames[k], mappedAttName) == 0)
					{
						fctx->relsAttNumsAtTupleSet[i][j] = k + 1;
						break;
					}
				}
			}
		}
	}

	/*
	 * Here we create a sorted relsAttNumsAtTupleSet by a key of the order of
	 * the tuple set attribute positions per relation
	 */
	int		 ***sortedRelsAttNumsAtTupleSet
	= (int ***) palloc(fctx->numRels * sizeof(int **));

	{
		int			min,
					minAttNum = -1,
					curr;

		for (i = 0; i < fctx->numRels; i++)
		{
			sortedRelsAttNumsAtTupleSet[i] = (int **) palloc(fctx->relsNAtt[i] * sizeof(int *));
			curr = -1;
			for (j = 0; j < fctx->relsNAtt[i]; j++)
			{
				min = MaxHeapAttributeNumber;
				for (k = 0; k < fctx->relsNAtt[i]; k++)
				{
					if ((fctx->relsAttNumsAtTupleSet[i][k] > curr) &&
						(fctx->relsAttNumsAtTupleSet[i][k] < min))
					{
						min = fctx->relsAttNumsAtTupleSet[i][k];
						minAttNum = k;
					}
				}
				curr = min;
				sortedRelsAttNumsAtTupleSet[i][j] = (int *) palloc(2 * sizeof(int));
				/* tuple set att num */
				sortedRelsAttNumsAtTupleSet[i][j][0] = min;
				/* matching rel att num */
				sortedRelsAttNumsAtTupleSet[i][j][1] = minAttNum;
			}
		}
		fctx->sortedRelsAttNumsAtTupleSet = sortedRelsAttNumsAtTupleSet;
	}

	/*
	 * here we find the proper casting for each attribute in each relation so
	 * that the equality function will be of same type. Also we setup the
	 * attribute types for the tuple set and setup the equality functions we
	 * are going to use.
	 */
	{
		/* tuple set attribute types (oid types) */
		fctx->tupleSetAttTypes = (Oid *) palloc(fctx->tupleSetNatt * sizeof(Oid));

		/*
		 * the first attribute is an internal fixed type (bytea, no need to
		 * specify it)
		 */
		fctx->tupleSetAttTypes[LABELS] = (Oid) NULL;
		/* tuple set per attribute proper equality functions */
		fctx->tupleSetAttEQFunctions
			= (TypeCacheEntry **) palloc(fctx->tupleSetNatt * sizeof(TypeCacheEntry *));
		/* no need for an equality function for the lables slot */
		fctx->tupleSetAttEQFunctions[LABELS] = NULL;

		/*
		 * a pointer to the attribute we are currently dealing with per
		 * relation.
		 */
		int			relsAttPointer[fctx->numRels];

		/*
		 * a structure were each slot is a relation id for the set of
		 * attribute with the same name we are currently dealing with. Not all
		 * slots are filled, in any one time we deal with the number of slots
		 * as the number of attributes with the same name shared among the
		 * relations
		 */
		int			sharedAttRelIDs[fctx->numRels];

		/*
		 * same as last but contains the original attribute number in the
		 * relsAttType structure
		 */
		int			sharedAttAttNum[fctx->numRels];

		/*
		 * num of attributes with the same name that can cast to the type of
		 * the slot relID
		 */
		int			numCanCastToAttType[fctx->numRels];

		/*
		 * number of attribute from the above that must be cast, i.e., have a
		 * different attribute type than the slot
		 */
		int			numMustCastToAttType[fctx->numRels];
		Oid			stype;
		Oid			ttype;
		int			numSharedRelIDs = 0;
		bool		simpleCast;
		int			minMustCast;
		int			minMustCastSharedRelID = -1;

		for (i = 0; i < fctx->numRels; i++)
			relsAttPointer[i] = -1;
		/* we do the casting checks per tuple set attribute */
		for (i = FIRSTATT; i < fctx->tupleSetNatt; i++)
		{
			numSharedRelIDs = 0;

			/*
			 * here we locate the attribute in each relation with the same
			 * name and put it in sharedAtt*
			 */
			for (j = 0; j < fctx->numRels; j++)
			{
				if ((relsAttPointer[j] + 1) >= fctx->relsNAtt[j])
					relsAttPointer[j] = -2;
				else if ((relsAttPointer[j] > -2) && (i == sortedRelsAttNumsAtTupleSet[j][relsAttPointer[j] + 1][0]))
				{
					relsAttPointer[j]++;
					sharedAttAttNum[numSharedRelIDs] = sortedRelsAttNumsAtTupleSet[j][relsAttPointer[j]][1];
					sharedAttRelIDs[numSharedRelIDs] = j;
					numSharedRelIDs++;
				}
			}
			/* here we find who can cast to whom */
			for (j = 0; j < numSharedRelIDs; j++)
			{
				numCanCastToAttType[j] = 0;
				numMustCastToAttType[j] = 0;
				stype = fctx->relsAttTypes[sharedAttRelIDs[j]][sharedAttAttNum[j]];
				for (k = 0; k < numSharedRelIDs; k++)
				{
					ttype = fctx->relsAttTypes[sharedAttRelIDs[k]][sharedAttAttNum[k]];
					if ((j != k) && (can_coerce_type(1, &stype, &ttype, COERCION_IMPLICIT)))
					{
						numCanCastToAttType[j]++;
						if (stype != ttype)
							numMustCastToAttType[j]++;
					}
				}
			}
			simpleCast = false;
			minMustCast = MaxHeapAttributeNumber;

			/*
			 * here we find the minimal number of casts absolutely neccessary
			 * in the queries, however, a future policy might consider casting
			 * smallest size in pages of relations or those with the most
			 * edges in the scheme graph to other relations
			 */
			for (j = 0; j < numSharedRelIDs; j++)
			{
				if (numCanCastToAttType[j] == (numSharedRelIDs - 1))
				{
					simpleCast = true;
					if (numMustCastToAttType[j] < minMustCast)
						minMustCast = numMustCastToAttType[j];
					minMustCastSharedRelID = j;
				}
			}

			/*
			 * if we can't cast all the attributes with the same name to the
			 * same type we go out in error. However, sometimes there are
			 * equality functions that can check for this but are difficult to
			 * identify. Therefore we give the user the option to explicity
			 * tell us that this attribute indeed can be casted. He does that
			 * by first handing out to the FD function an already explicity
			 * casted relations
			 */
			if (!simpleCast)
              ereport(ERROR,
               (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                errmsg("Casting problem!"),
                errdetail("Same names at different relations but with differnt types\
			 that are difficult to cast.You can use CAST to explicitly change the attributes\
		to the same type.The name of the last problematic attribute is:%s ",
					 fctx->resultTupleAttNames[i - 1])));
			else
				fctx->tupleSetAttTypes[i]
					= fctx->relsAttTypes[sharedAttRelIDs[minMustCastSharedRelID]][sharedAttAttNum[minMustCastSharedRelID]];
		}
        
        /* here we get the equality function for same types for each attribute in the result tuple.*/
		for (i = 0; i < fctx->resultTupleNAtt; i++)
			fctx->tupleSetAttEQFunctions[i + 1]
				= getEQFunction(fctx->tupleSetAttTypes[i + 1]);
	}
    
	/* Here we create the scheme graph */
	{
		int			k,
					z;

		for (i = 0; i < fctx->numRels; i++)
			fctx->bits_scheme_graph[i] = 0;
		for (i = 0; i < fctx->numRels; i++)
		{
			for (j = i + 1; j < fctx->numRels; j++)
			{
				k = 0;
				z = 0;
				while ((k < fctx->relsNAtt[i]) && (z < fctx->relsNAtt[j]))
				{
					if (sortedRelsAttNumsAtTupleSet[i][k][0] ==
						sortedRelsAttNumsAtTupleSet[j][z][0])
					{
						fctx->bits_scheme_graph[i] = (fctx->bits_scheme_graph[i] | nthBitFromLeft(j));
						fctx->bits_scheme_graph[j] = (fctx->bits_scheme_graph[j] | nthBitFromLeft(i));
						break;
					}
					else if (sortedRelsAttNumsAtTupleSet[i][k][0] <
							 sortedRelsAttNumsAtTupleSet[j][z][0])
						k++;
					else
						z++;
				}
			}
		}
	}
    
	/* Lets check that the scheme graph is a connected scheme graph.
     * Because, the FD is only useful on connected scheme graphs.
     * If it is not, we must exit.*/
	{
		RBITS		relsMask = (~0x0);

		relsMask >>= fctx->numRels;
		relsMask = ~relsMask;
		RBITS		repository = nthBitFromLeft(0);

		repository |= fctx->bits_scheme_graph[0];
		RBITS		wereChecked = nthBitFromLeft(0);
		bool		wasAddition = false;

		for (i = 0; i < fctx->numRels; i++)
		{
			wasAddition = false;
			for (j = 1; j < fctx->numRels; j++)
			{
				if ((!(varNthBitFromLeft(wereChecked, j))) && (varNthBitFromLeft(repository, j)))
				{
					repository |= fctx->bits_scheme_graph[j];
					varSetNthBitFromLeft(repository, j);
					varSetNthBitFromLeft(wereChecked, j);
					wasAddition = true;
					if (repository == relsMask)
						break;
				}
			}
			if ((!wasAddition) || (repository == relsMask))
				break;
		}
		if (repository != relsMask)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("Disconnected components detected!"),
			  errhint("The relations must be connected to each other! for example A=(attribute1,attribute2), B=(attribute2,attribute3). A and B are connected because they share a common attribute2. Let C=(attribute4,attribute5). C is not connected to either A or B thus, the relations are not connected to each other. Make sure they do before you run the algorithm again.")));
	}
    
	/* lets generate the result tuple, tuple description. */
	{
		TupleDesc	desc = CreateTemplateTupleDesc(fctx->resultTupleNAtt, false);
		TupleDesc	tupleSetDesc = CreateTemplateTupleDesc(NUMATT, false);

        /* This is the added labels part of a tuple set. A tuple set is composed of a set
         * of lables that represent the tuples from different relations in the set. And of
         * course the attributes of these tuples. Since the attributes are the result of
         * the join of these tuples, we do not need to repeat columns or actually keep
         * the tuples in their original form. */
		TupleDescInitEntry(tupleSetDesc, LABELS_ALIGNED, "participatingRelationsOIDs"
						   ,BYTEAOID, -1, 0);
		for (i = 0; i < fctx->resultTupleNAtt; i++)
			if (!is_array_type(fctx->tupleSetAttTypes[i + 1]))
			{
				TupleDescInitEntry(desc, (i + 1), fctx->resultTupleAttNames[i]
								   ,fctx->tupleSetAttTypes[i + 1], -1, 0);
				TupleDescInitEntry(tupleSetDesc, (i + 2), fctx->resultTupleAttNames[i]
								   ,fctx->tupleSetAttTypes[i + 1], -1, 0);
			}
			else
			{
				TupleDescInitEntry(desc, (i + 1), fctx->resultTupleAttNames[i]
								   ,fctx->tupleSetAttTypes[i + 1], -1, 1);

				/*
				 * if its an array then lets give it 3 dim will occur in any
				 * array. more than we will need to support at later
				 */
				TupleDescInitEntry(tupleSetDesc, (i + 2), fctx->resultTupleAttNames[i]
								   ,fctx->tupleSetAttTypes[i + 1], -1, 1);
			}
		fctx->resultTupleDesc = desc;
		fctx->tupleSetDesc = tupleSetDesc;
		/* HERE WE REGISTER THE OUTPUT SCHEME!!! */
	}
}

/* Here we initialize the SPI structures that will allow us to read the relations.
 * Basically, we open them once here and iterate on them by moving the cursor to the
 * start of the portal each time without reopening them. The use of cursors is important
 * because of transactional issues where because the FD is a long operation we must always
 * work with the same data and thus, we must use the same initial snapshot. Basically opening
 * a read only cursor should do the trick */
void
initialize_SPI_structures(PG_FUNCTION_ARGS, alg_fctx * fctx, bool noDuplicatesTest
						  ,FuncCallContext *funcctx)
{
	SPI_connect();
	fctx->portals = (Portal *) palloc(fctx->numRels * sizeof(Portal));
	fctx->plans = (void **) palloc(fctx->numRels * sizeof(void *));
	int			i = 0;

	{
		/* checking if there are duplicates in the relations. Only if the user requested. */
		if (noDuplicatesTest)
		{
			bool		noDuplicates = true;

			for (i = 0; i < fctx->numRels; i++)
			{
				char		query[255];
				char	   *to = query;

				to = (char *) stpcpy(to, "SELECT 'a' FROM pg_class WHERE relname='");
				to = (char *) stpcpy(to, get_rel_name(fctx->relsOid[i]));
				to = (char *) stpcpy(to, "' and relhaspkey='t' and relnamespace='");
				char		charNameSpaceOID[20];

				sprintf(charNameSpaceOID, "%d", get_rel_namespace(fctx->relsOid[i]));
				to = (char *) stpcpy(to, charNameSpaceOID);
				to = (char *) stpcpy(to, "'");
				void	   *plan;
				Portal		portal;

				if ((plan = SPI_prepare(query, 0, NULL)) == NULL)
                 elog(ERROR, "At %s(%d) - Could not initialize SPI structures as SPI_prepare (%s) returned NULL",
                     __FILE__, __LINE__, query);
				if ((portal
					 = SPI_cursor_open(NULL, plan, NULL, NULL, true)) == NULL)
                 elog(ERROR, 
                     "At %s(%d) - Could not initialize SPI structures as SPI_cursor_open (%s) returned NULL", 
                     __FILE__, __LINE__, query);
				SPI_cursor_fetch(portal, true, 1);
				if (SPI_processed <= 0)
					noDuplicates = false;
				else
					SPI_freetuptable(SPI_tuptable);
				SPI_cursor_close(portal);
				SPI_freeplan(plan);
				if (!noDuplicates)
				{
					to = query;
					to = (char *) stpcpy(to, "SELECT 'a' FROM ");
					to = (char *) stpcpy(to, get_namespace_name(get_rel_namespace(fctx->relsOid[i])));
					to = (char *) stpcpy(to, ".");
					to = (char *) stpcpy(to, get_rel_name(fctx->relsOid[i]));
					to = (char *) stpcpy(to, " HAVING count(*)>1");
					if ((plan = SPI_prepare(query, 0, NULL)) == NULL)
                     elog(ERROR, 
                         "At %s(%d) - Could not initialize SPI structures as SPI_prepare (%s) returned NULL",
                          __FILE__, __LINE__, query);
					if ((portal
					= SPI_cursor_open(NULL, plan, NULL, NULL, true)) == NULL)
                       elog(ERROR, 
                        "At %s(%d) - Could not initialize SPI structures as SPI_cursor_open (%s) returned NULL", 
                         __FILE__, __LINE__, query);
					SPI_cursor_fetch(portal, true, 1);
					if (SPI_processed <= 0)
						noDuplicates = false;
					else
						noDuplicates = true;
					SPI_freetuptable(SPI_tuptable);
					SPI_cursor_close(portal);
					SPI_freeplan(plan);
				}
				if (!noDuplicates)
					break;
			}
			fctx->noDuplicates = noDuplicates;
		}
	}

    /* Here we actually setup the query and open the cursors. We also handle
     * the proper castings and formulate a different query for the event there
     * are duplicates. */
	{
		fctx->queries = (char **) palloc(fctx->numRels * sizeof(char *));
		int			j,
					k;
		char	   *query;
		char	   *to;
		int			querySize;
		bool		toFreeQuery;
		bool		firstAtt = true;

		for (i = 0; i < fctx->numRels; i++)
		{
			querySize = 1024;
			fctx->queries[i] = (char *) palloc(querySize * sizeof(char));
			query = fctx->queries[i];
			to = query;
			int			iterateNoDuplicates = 1;

			if (fctx->noDuplicates)
				query = addAndExpandStringBy(query, &querySize, "SELECT "
											 ,querySize, &to, &toFreeQuery);
			else
			{
				query = addAndExpandStringBy(query, &querySize, "SELECT DISTINCT ON ("
											 ,querySize, &to, &toFreeQuery);
				iterateNoDuplicates = 2;
			}
			bool		closed = false;

			for (k = 0; k < iterateNoDuplicates; k++)
			{
				firstAtt = true;
				for (j = 0; j < fctx->relsNAtt[i]; j++)
				{
					if (!firstAtt)
					{
						query = addAndExpandStringBy(query, &querySize, ","
											  ,querySize, &to, &toFreeQuery);
					}
					firstAtt = false;
					query = addAndExpandStringBy(query, &querySize,
												 (char *) get_attname(fctx->relsOid[i], fctx->relsValidAttNumAtRelation[i][j])
											  ,querySize, &to, &toFreeQuery);
					if (fctx->relsAttTypes[i][j] != fctx->tupleSetAttTypes[fctx->relsAttNumsAtTupleSet[i][j]])
					{
						query = addAndExpandStringBy(query, &querySize, "::"
											  ,querySize, &to, &toFreeQuery);
						query = addAndExpandStringBy(query, &querySize,
													 format_type_with_typemod(fctx->tupleSetAttTypes[fctx->relsAttNumsAtTupleSet[i][j]], -1)
											  ,querySize, &to, &toFreeQuery);
					}
				}
				if (!fctx->noDuplicates && !closed)
				{
					query = addAndExpandStringBy(query, &querySize, ") "
											  ,querySize, &to, &toFreeQuery);
					closed = true;
				}
			}
            /* CTID is used as the label of a tuple (together with the table id) thus,
             * we must have it */
			query = addAndExpandStringBy(query, &querySize, ",CTID FROM "
										 ,querySize, &to, &toFreeQuery);
			query = addAndExpandStringBy(query, &querySize, get_namespace_name(get_rel_namespace(fctx->relsOid[i]))
										 ,querySize, &to, &toFreeQuery);
			query = addAndExpandStringBy(query, &querySize, "."
										 ,querySize, &to, &toFreeQuery);
			query = addAndExpandStringBy(query, &querySize, get_rel_name(fctx->relsOid[i])
										 ,querySize, &to, &toFreeQuery);
			if ((fctx->plans[i] = SPI_prepare(query, 0, NULL)) == NULL)
               elog(ERROR, 
                   "At %s(%d) - Could not initialize SPI structures as SPI_prepare (%s) returned NULL",
                    __FILE__, __LINE__, query);
			if ((fctx->portals[i]
				 = SPI_cursor_open(NULL, fctx->plans[i], NULL, NULL, true)) == NULL)

				/*
				 * note that read_only in cursor_open must be true otherwise
				 * ctid can change.
				 */
               elog(ERROR, 
                "At %s(%d) - Could not initialize SPI structures as SPI_cursor_open (%s) returned NULL", 
                 __FILE__, __LINE__, query);
		}
	}
}
