/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#ifndef ALGSTRUCTS_H
#define ALGSTRUCTS_H

#include "postgres.h"
#include "utils/palloc.h"
#include "funcapi.h"
#include "fmgr.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"
#include "executor/spi_priv.h"
#include "utils/lsyscache.h"
#include "utils/typcache.h"
#include "queues.h"

/* for compatibility with Postgres <= 8.2 */
#ifndef SET_VARSIZE
# define SET_VARSIZE(PTR,SIZE) (VARATT_SIZEP(PTR) = (SIZE))
#endif

/* a safety measure if we do not calculate the max size of the result tuple */
#define RESULT_TUPLE_ASSUMED_MAX_SIZE_IN_PAGES 1
#define RESULT_TUPLE_ASSUMED_MAX_SIZE_IN_BYTES (RESULT_TUPLE_ASSUMED_MAX_SIZE_IN_PAGES*BLCKSZ)

/* Mapping from a relation attribute to an alias */
typedef struct
{
	int			resultTuplePotnetialAttNamesID;
	char	   *relAttName;
}	Mapping;

/* a dictionary of mappings */
typedef struct
{
    /* list of unique aliases.*/
	char	  **resultTuplePotentialAttNames;
	int			numPotentialAttNames;
    /* Holds a list of mappings for each relation. */
	Mapping  ***relsAttsMappings;
	int		   *numMappings;
}	Dictionary;

/* a persistant state strucutre between calls to the OnDemand Procedure. */
typedef struct
{
	/* index in Q */
	int			q;
	/* index in P */
	int			p;
	memQueue   *Q;
    /* pointer in Q */
	memQueueNode *qPtr;
    /* current relation tuple we work with */
	DTuple		P;
	TupleDesc	Pdesc;
	int			PrelID;
    /* we keep the last SPI (P) tuptable so we can free it properly. */
	SPITupleTable *lastTuptable;
    /* current relation we are scanning */
	int			cur_rel;
    /* The portals of all the relations. They are the same as in the SPI structures in algstructs
     * but we need to scan them seperately here with a seperate cursor. */
	Portal	   *portals;
}	OnDemandState;

/* This is the structure of the adaptive upper bound on how many tuple sets we are
 * going to extend at once. */
typedef struct
{
	double		previousScore;
	int			nextUpperBound;
	int			stage;
	int			maxInputedExtendTuples;
	int			actualInputedExtendTuples;
}	adaptive_extend_upperbound;

/* This is the general context of the algorithm. 
 * Most of the algorithms structures are here. */
typedef struct
{
    /* number of arguments in the input */
	int			nargs;
    /* number of inputted relations */
	int			numRels;
    /* a portal for each relation */
	Portal	   *portals;
    /* a query for each relation */
	char	  **queries;
    /* a plan for each relation */
	void	  **plans;
	/* relations oid's */
	Oid		   *relsOid;
	/* the size in pages of each relation. */
	int		   *relsPages;
	/* true if relation has an index, false otherwise. */
	bool	   *relsHasIndex;

	/*
	 * the array will contain RELIDS where cell 0 will contain the smallest
	 * paged size relation ID.
	 */
	int		   *sortedRelsByRelsPages;
    /* the estimated or real maximum tuple size for each relation */
	int		   *relsMaxTupleSize;
	/* number of attributes of each relation. */
	int		   *relsNAtt;

	/*
	 * for each relation and each of its attributes, specifies the attribute
	 * number at the result tuple.
	 */
	int		  **relsAttNumsAtTupleSet;
	/* same as last but sorted by tuple set att num key. */
	int		 ***sortedRelsAttNumsAtTupleSet;
    /* does not include dropped columns of relations */
	int		  **relsValidAttNumAtRelation;
	/* for each relation lets save its attributes types; */
	Oid		  **relsAttTypes;
	/* The result tuple attribute names. */
	char	  **resultTupleAttNames;
	/* The result tupl attribute types. */
	Oid		   *tupleSetAttTypes;
    /* equality functions for same types of the respected tuple set type */
	TypeCacheEntry **tupleSetAttEQFunctions;
    /* tuple description of the result tuple */
	TupleDesc	resultTupleDesc;
	/* The result tuple number of attributes. */
	int			resultTupleNAtt;
	/* in bytes. how much its HeapTuple occuppies in memory. */
	int			tupleSetMaxSize;
	/* how many result tuples can fit in one page. */
	int			m_NumResultTuplesInOnePage;
    
    /* indicate if we use the default value */
	bool		assumingResultTupleMaxSize;
    /* The tuple set description (used to dumb tuple set to storage) */
	TupleDesc	tupleSetDesc;
    /* tuple set number of attributes */
	int			tupleSetNatt;
    /* dictionary of inputted mappings */
	Dictionary *dictionary;
	/* 0 not connected, 1 connected. */
	RBITS	   *bits_scheme_graph;
    /* tuple store of outputted result tuples */
	storageQueue complete;
    /* tuple store of not yet outputted result tuples */
	storageQueue precomplete;
    /* indicates if there is a need to check for duplicates and later after testing
     * if there are in fact duplicates. */
	bool		noDuplicates;
    /* The algorithm state in the state machine. */
	int			algorithm_state;
    /* a queue for reading from virtual incomplete (precomplete with the relations) */
	memQueue	Q;
    /* a queue for extending */
	memQueue	E;
	/* a pointer to the last sent tset for printing or pipelining. */
	memQueueNode *printedTSet;

	/*
	 * if you are using queue to print then this signifies that we started
	 * printing when we come back from backend.
	 */
	bool		startedPrintingQueue;
    /* pointer to the printed tuple */
	HeapTuple	printedTupleAddress;
    /* indicates if we want to use indices in the relations for extending */
	bool		noIndexScans;
	double		startTime;
	double		endTime;
    /* the persistent state for the on demand procedure */
	OnDemandState *onDemandState;
    /* the MB input parameter */
	int			MB;
    /* the policy used for extending - inputted */
	int			extendingPolicy;
    /* the last relation we have scanned in extending */
	int			extendingLastRelScanned;
    /* statistical variable */
	int			extendingNaivePolicyNumCompleteScans;
    /* used for extending with complex policies */
	memQueueNode *policyWorkNode;
	RBITS		policySequence;
	RBITS		relationsToScan;
    /* query for extending when using indices */
	char	   *query;
	int			querySize;
	Datum	   *queryParams;
	Oid		   *queryParamsTypes;
	int			queryParamsSize;
    /* future indicator, not really used currently. */
	bool		trueIndexFalseNonIndexBasedAlgorithm;
    /* adaptive upper bound for extending structures */
	adaptive_extend_upperbound *dynamic_extend_cap;
    /* an additional memory context to handle accesses to the tuple store.
     * Needed to prevent leaking and all kinds of weird issues */
	MemoryContext per_query_ctx;
}	alg_fctx;

extern Oid *parseOIDs(text *relsStr, int *numOfRels);
extern Oid *parseOIDs_(char *relsStrCp, int *numOfRels);
extern void initialize_general_structures(PG_FUNCTION_ARGS, alg_fctx * fctx, FuncCallContext *funcctx);
extern void initialize_general_structures_(PG_FUNCTION_ARGS, alg_fctx * fctx, bool parsing, FuncCallContext *funcctx);
extern void initialize_SPI_structures(PG_FUNCTION_ARGS, alg_fctx * fctx, bool noDuplicatesTest
						  ,FuncCallContext *funcctx);

#endif   /* ALGSTRUCTS_H */
