/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#include "postgres.h"
#include "algutils.h"
#include "catalog/heap.h"
#include "catalog/catversion.h"
#include "tset.h"
#include "tsetfuncs.h"
#include "queues.h"
#include "queuesfuncs.h"
#include "utils/datum.h"
#include <stdarg.h>

/* Just copying a datum */
Datum
myDatumCopy(Datum value, bool typByVal, int typLen)
{
	Datum		res;

	if (typByVal)
		res = value;
	else
	{
		Size		realSize;
		char	   *s;

		if (DatumGetPointer(value) == NULL)
			return PointerGetDatum(NULL);
		realSize = datumGetSize(value, typByVal, typLen);
		s = (char *) palloc(realSize);
		memcpy(s, DatumGetPointer(value), realSize);
		res = PointerGetDatum(s);
	}
	return res;
}

/* Freeing a datum */
void
myDatumFree(Datum value, bool typByVal, int typLen)
{
	if (!typByVal)
	{
		Pointer		s = DatumGetPointer(value);

		pfree(s);
	}
}

/* sorting an array of Oids */
void
sortOidArray(Oid *array, int numElements)
{
	int			swapint,
				i,
				j;

	for (i = 0; i < (numElements - 1); i++)
		for (j = (i + 1); j < numElements; j++)
			if (array[i] > array[j])
			{
				swapint = array[i];
				array[i] = array[j];
				array[j] = swapint;
			}
}

/* updating the max tuple size */
void
updateMaxTupleSize(alg_fctx * fctx, int relID, HeapTuple tuple)
{
	int			size = tupleSize(tuple);

	if (fctx->relsMaxTupleSize[relID] < size)
		fctx->relsMaxTupleSize[relID] = size;
}

/* updating the max result tuple size*/
void
updateResultTupleMaxSize(alg_fctx * fctx)
{
	int			tmpSize = 0;
	int			i = 0;

	for (i = 0; i < fctx->numRels; i++)
		tmpSize += fctx->relsMaxTupleSize[i];
	fctx->tupleSetMaxSize = tmpSize;
	double		blcksz = BLCKSZ;

	if (tmpSize > 0)
		fctx->m_NumResultTuplesInOnePage = (int) (blcksz / (double) tmpSize);
	else
		fctx->m_NumResultTuplesInOnePage = 1;
	if (fctx->m_NumResultTuplesInOnePage < 1)
		fctx->m_NumResultTuplesInOnePage = 1;
}

/* getting an equality function for same type */
TypeCacheEntry *
getEQFunction(Oid typeid)
{
	TypeCacheEntry *typentry;

	/*
	 * Find the data type in the typcache, and ask for eq_opr info.
	 */
	typentry = lookup_type_cache(typeid, TYPECACHE_EQ_OPR_FINFO);
	if (!OidIsValid(typentry->eq_opr_finfo.fn_oid))
		ereport(ERROR, (errcode(ERRCODE_UNDEFINED_FUNCTION),
				errmsg("could not identify an equality operator for type %s",
					   format_type_be(typeid))));
	return typentry;
}

/* auto expanding a datum array in memory */
Datum *
getAndExpandDatumArray(Datum *oldDatumArr, int oldDatumArraySize, int newDatumArraySize)
{
	if (oldDatumArr == NULL)
		return (Datum *) palloc(newDatumArraySize * sizeof(Datum));
	else
	{
		if (oldDatumArraySize < newDatumArraySize)
		{
			Datum	   *newArr = (Datum *) palloc(newDatumArraySize * sizeof(Datum));

			memcpy(newArr, oldDatumArr, oldDatumArraySize * sizeof(Datum));
			return newArr;
		}
	}
	return oldDatumArr;
}

/* auto expanding an Oid array in memory */
Oid *
getAndExpandOidArray(Oid *oldOidArr, int oldOidArraySize, int newOidArraySize)
{
	if (oldOidArr == NULL)
		return (Oid *) palloc(newOidArraySize * sizeof(Oid));
	else
	{
		if (oldOidArraySize < newOidArraySize)
		{
			Oid		   *newArr = (Oid *) palloc(newOidArraySize * sizeof(Oid));

			memcpy(newArr, oldOidArr, oldOidArraySize * sizeof(Oid));
			return newArr;
		}
	}
	return oldOidArr;
}

/* auto expanding a string array in memory */
char *
addAndExpandStringBy(char *oldStr, int *oldStrAllocSize, char *addStr, int sizeToAdd
					 ,char **relative_ptr, bool *free_old_str)
{
	int			oldStrSize = (int) *relative_ptr - (int) oldStr;
	int			addStrSize = strlen(addStr);

	/* +1 for null at the end */
	if (*oldStrAllocSize >= (oldStrSize + addStrSize + 1))
	{
		*relative_ptr = (char *) stpcpy(*relative_ptr, addStr);
		return oldStr;
	}
	*oldStrAllocSize = *oldStrAllocSize + addStrSize + sizeToAdd * sizeof(char);
	char	   *newStr = (char *) palloc(*oldStrAllocSize);
	char	   *to = newStr;

	to = (char *) stpcpy(to, oldStr);
	to = (char *) stpcpy(to, addStr);
	if (relative_ptr != NULL)
		*relative_ptr = to;
	
    Assert(relative_ptr != NULL);
    
	if (*free_old_str)
		pfree(oldStr);
	*free_old_str = true;
	return newStr;
}

/* checks if an attribute in the result tuple also appear in the relation specified
 * in the relID */
bool
isResultTupleAttOccursInRelID(alg_fctx * fctx, int resultTupleAttID, int relID)
{
	int			i = 0;

	for (i = 0; i < fctx->relsNAtt[relID]; i++)
	{
		if (fctx->relsAttNumsAtTupleSet[relID][i] == resultTupleAttID)
			return true;
	}
	return false;
}

/*
 *	in some RelID of a tuple set of course. Where the tuple set is represented
 *	by the relsID.
 */
bool
isResultTupleAttOccursInSomeRelID(alg_fctx * fctx, int resultTupleAttID, bytea *relsID)
{
	int			numIDs = numOfTuplesInTSet(fctx, relsID);
	int			j = 0;

	for (j = 0; j < numIDs; j++)
		if (isResultTupleAttOccursInRelID(fctx, resultTupleAttID
								,getTableIdOfRelsOidInTSet(fctx, relsID, j)))
			return true;
	return false;
}

/* Same as isResultTupleAttOccursInSomeRelID but using a BITS array instead of a bytea set
 * of lables. */
bool
isResultTupleAttOccursInSomeRelIDWithSpecifiedRelations(alg_fctx * fctx, int resultTupleAttID,
													  RBITS relationsToCheck)
{
	int			i = 0,
				j = 0;
	RBITS		localRelationsToCheck = relationsToCheck;

	while (localRelationsToCheck)
	{
		if (varNthBitFromLeft(localRelationsToCheck, 0))
			for (j = 0; j < fctx->relsNAtt[i]; j++)
			{
				if (fctx->relsAttNumsAtTupleSet[i][j] == resultTupleAttID)
					return true;
			}
		i++;
		localRelationsToCheck <<= 1;
	}
	return false;
}

/*
 * A policy to choose the next relation to extend to.
 * majority is to extend to the relation most tuple set currently needs
 * to extend to.
 * -1 means, no relation was chosen
 */
int
policy_majority_Q_one_tuple_set_at_a_time(alg_fctx * fctx, memQueue * Q, int tsetsNum)
{
	int			j = 0,
				maxSum = 0,
				i = 0;
	RBITS		tmp = 0,
				tmpRels = 0;
	int			relsSum[fctx->numRels];

	for (j = 0; j < fctx->numRels; j++)
		relsSum[j] = 0;
	memQueueNode *ptr = NULL;

	if (fctx->policyWorkNode)
		ptr = fctx->policyWorkNode;
	else
	{
		ptr = getMemQueueFirstElementPtr(Q);
		if (!(ptr))
			return -1;
	}
	tmp = ptr->toVisit;

    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	ptr = getMemQueueFirstElementPtr(Q);
	while (ptr != NULL)
	{
		tmpRels = ((ptr->toVisit) & tmp);
		j = 0;
		while (tmpRels)
		{
			varNthBitFromLeft(tmpRels, 0) && relsSum[j]++;
			tmpRels <<= 1;
			j++;
		}
		ptr = getMemQueueNextElementPtr(ptr);
	}
	for (j = 0; j < fctx->numRels; j++)
		if (relsSum[j] > maxSum)
		{
			i = j;
			maxSum = relsSum[j];
		}
	return i;
}

/*
 * same as with policy_majority_Q_one_tuple_set_at_a_time but with sequencing.
 * -1 means, no relation was chosen
 */
int
policy_majority_Q_one_tuple_set_at_a_time_with_sequencing(alg_fctx * fctx, memQueue * Q,
														  int tsetsNum)
{
	int			j = 0,
				maxSum = 0,
				i = 0;
	int			relsSum[fctx->numRels];

	for (j = 0; j < fctx->numRels; j++)
		relsSum[j] = 0;
	RBITS		tmp = 0;
	RBITS		tmpRels = 0;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	bool		wasAddToSum = false;
	memQueueNode *ptr = NULL;

	if (fctx->policyWorkNode)
	{
		ptr = fctx->policyWorkNode;
		tmp = ptr->toVisit;
	}
	else
	{
		ptr = getMemQueueFirstElementPtr(Q);
		if (!(ptr))
			return -1;
		if (fctx->policySequence)
		{
			tmp = fctx->policySequence;
		}
		else
		{
			fctx->policySequence = relsMask;
			tmp = ptr->toVisit;
            
			Assert(tmp);

		}
	}
	
    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	ptr = getMemQueueFirstElementPtr(Q);
	while (ptr != NULL)
	{
		tmpRels = ((ptr->toVisit) & tmp);
		j = 0;
		while (tmpRels)
		{
			if (varNthBitFromLeft(tmpRels, 0))
				relsSum[j]++;
			if (varNthBitFromLeft(tmpRels, 0))
				wasAddToSum = true;
			tmpRels <<= 1;
			j++;
		}
		ptr = getMemQueueNextElementPtr(ptr);
	}
	if (wasAddToSum)
	{
		for (j = 0; j < fctx->numRels; j++)
			if (relsSum[j] > maxSum)
			{
				i = j;
				maxSum = relsSum[j];
			}
		varUnSetNthBitFromLeft(fctx->policySequence, i);
		return i;
	}
	else
	{
		fctx->policySequence = 0;
		fctx->policyWorkNode = NULL;
		return policy_majority_Q_one_tuple_set_at_a_time_with_sequencing(fctx, Q, tsetsNum);
	}
}

/*
 * mincost policy choose the relations the tuple sets needs to extend to but that has
 * the least pages in it.
 * -1 means, no relation was chosen
 */
int
policy_mincost_Q_one_tuple_set_at_a_time(alg_fctx * fctx, memQueue * Q, int tsetsNum)
{
	int			j = 0;
	RBITS		tmp = 0;
	memQueueNode *ptr = NULL;

	if (fctx->policyWorkNode)
		ptr = fctx->policyWorkNode;
	else
	{
		ptr = getMemQueueFirstElementPtr(Q);
		if (!(ptr))
			return -1;
	}
	tmp = ptr->toVisit;
	
    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	for (j = 0; j < fctx->numRels; j++)
		if (varNthBitFromLeft(tmp, fctx->sortedRelsByRelsPages[j]))
			return fctx->sortedRelsByRelsPages[j];
	return -1;
}

/*
 * same as one at a time, but with sequencing.
 * -1 means, no relation was chosen
 */
int
policy_mincost_Q_one_tuple_set_at_a_time_with_sequencing(alg_fctx * fctx, memQueue * Q,
														 int tsetsNum)
{
	int			j = 0,
				relID = 0;
	RBITS		tmp = 0;
	RBITS		tmpRels = 0;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	memQueueNode *ptr = NULL;
	RBITS		sequence = fctx->policySequence;

	if (fctx->policyWorkNode)
	{
		ptr = fctx->policyWorkNode;
		tmp = ptr->toVisit;
	}
	else
	{
		ptr = getMemQueueFirstElementPtr(Q);
		if (!(ptr))
			return -1;
		if (sequence)
		{
			for (j = 0; j < fctx->numRels; j++)
			{
				relID = fctx->sortedRelsByRelsPages[j];
				if (varNthBitFromLeft(sequence, relID))
				{
					tmp = nthBitFromLeft(relID);
					break;
				}
			}
			while (ptr != NULL)
			{
				tmpRels |= (ptr->toVisit);
				if (tmpRels & tmp)
				{
					varUnSetNthBitFromLeft(fctx->policySequence, relID);
					return relID;
				}
				if ((tmpRels & sequence) == sequence)
					break;
				ptr = getMemQueueNextElementPtr(ptr);
			}
			ptr = NULL;
			tmp = fctx->policySequence & tmpRels;
			if (!(tmp))
			{
				ptr = getMemQueueFirstElementPtr(Q);
				fctx->policySequence = relsMask;
				tmp = ptr->toVisit;
			}
		}
		else
		{
			fctx->policySequence = relsMask;
			tmp = ptr->toVisit;
		}
	}
	
    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	for (j = 0; j < fctx->numRels; j++)
		if (varNthBitFromLeft(tmp, fctx->sortedRelsByRelsPages[j]))
		{
			varUnSetNthBitFromLeft(fctx->policySequence, fctx->sortedRelsByRelsPages[j]);
			return fctx->sortedRelsByRelsPages[j];
		}
	return -1;
}

/*
 * random relation choice but only from those the tuple sets needs to extend to.
 * (the tuple sets in Q)
 * -1 means, no relation was chosen
 */
int
policy_random_Q(alg_fctx * fctx, memQueue * Q, int tsetsNum)
{
	int			i = 0;
	int			j = 0;
	int			numRelsToExtend = 0;
	RBITS		tmpRels = 0;
	RBITS		tmp;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	memQueueNode *ptr = getMemQueueFirstElementPtr(Q);

	while (ptr != NULL)
	{
		tmpRels |= (ptr->toVisit);
		if (tmpRels == relsMask)
			break;
		ptr = getMemQueueNextElementPtr(ptr);
	}
	tmp = tmpRels;
	while (tmp)
	{
		if (varNthBitFromLeft(tmp, 0))
			numRelsToExtend++;
		tmp <<= 1;
	}
	if (numRelsToExtend > 0)
	{
		int			k = 0;

		i = ((int) (numRelsToExtend * (rand() / (RAND_MAX + 1.0)))) + 1;
	    
        Assert(i <= numRelsToExtend);
        
		j = 0;
		while ((j < i) && (j < numRelsToExtend))
		{
			if (varNthBitFromLeft(tmpRels, k))
			{
				j++;
			}
			k++;
		}
		k--;
		return k;
	}
	return -1;
}

/*
 * naive policy. mostly useless.
 * -1 means, no relation was chosen
 */
int
policy_naive_Q(alg_fctx * fctx, memQueue * Q, int tsetsNum)
{
	if (fctx->extendingLastRelScanned < 0)
		return 0;
	if ((fctx->extendingLastRelScanned + 1) < fctx->numRels)
		return (fctx->extendingLastRelScanned + 1);
	fctx->extendingNaivePolicyNumCompleteScans++;
	if (fctx->extendingNaivePolicyNumCompleteScans == fctx->numRels)
		return -1;
	return 0;
}

/* naive policy, chooses the relation in a deterministic manner with no regard to the
 * tuple sets.*/
int
policy_naive(alg_fctx * fctx, int **checkTable, int tsetsNum)
{
	if (fctx->extendingLastRelScanned < 0)
		return 0;
	if ((fctx->extendingLastRelScanned + 1) < fctx->numRels)
		return (fctx->extendingLastRelScanned + 1);
	fctx->extendingNaivePolicyNumCompleteScans++;
	if (fctx->extendingNaivePolicyNumCompleteScans == fctx->numRels)
		return -1;
	return 0;
}

/* checks using a tree search the plantree if it contains a sequence scan. and return
 * true if it does.*/
bool
isPlanUsingSeqScan(Plan *planTree)
{
	if (nodeTag(planTree) == T_SeqScan)
		return true;
	if (nodeTag(planTree) == T_Append)
	{
		Append	   *appendplan = (Append *) planTree;
		ListCell   *lst;

		foreach(lst, appendplan->appendplans)
		{
			Plan	   *subnode = (Plan *) lfirst(lst);

			return (isPlanUsingSeqScan(subnode));
		}
	}
	if (nodeTag(planTree) == T_SubqueryScan)
	{
		SubqueryScan *subqueryscan = (SubqueryScan *) planTree;
		Plan	   *subnode = subqueryscan->subplan;

		return (isPlanUsingSeqScan(subnode));
	}
	if (innerPlan(planTree))
		if (isPlanUsingSeqScan(innerPlan(planTree)) == true)
			return true;
	if (outerPlan(planTree))
		if (isPlanUsingSeqScan(outerPlan(planTree)) == true)
			return true;
	return false;
}

/*simply recombines queues using their head and tail pointers.*/
void
recombineQueues(memQueue * destQ, int numQueues,...)
{
	memQueue   *sourceQ = NULL;
	memQueue	destQtmp;

	_initializeMemQueue(&destQtmp, destQ->allocatedBytes);
	int			i;
	va_list		ap;

	va_start(ap, numQueues);
	for (i = 0; i < numQueues; i++)
	{
		sourceQ = va_arg(ap, memQueue *);
		if (sourceQ->head != NULL)
		{
			if (destQtmp.head == NULL)
				destQtmp.head = sourceQ->head;
			else
				destQtmp.tail->next = (struct memQueueNode *) sourceQ->head;
			destQtmp.numElements += sourceQ->numElements;
			destQtmp.sizeInBytes += sourceQ->sizeInBytes;
			destQtmp.tail = sourceQ->tail;
		}
	}
	_initializeMemQueue(destQ, destQ->allocatedBytes);
	destQ->head = destQtmp.head;
	destQ->tail = destQtmp.tail;
	destQ->numElements = destQtmp.numElements;
	destQ->sizeInBytes = destQtmp.sizeInBytes;
	va_end(ap);
}

/*
 * a random policy to choose a relation. However, only those that the tuple sets in Q needs
 * to extend to.
 * -1 means, no relation was chosen
 */
int
policy_random_Q_one_tuple_set_at_a_time(alg_fctx * fctx, memQueue * Q, int tsetsNum)
{
	int			i = 0;
	int			j = 0;
	int			numRelsToExtend = 0;
	RBITS		tmp;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	memQueueNode *ptr = NULL;

	if (fctx->policyWorkNode)
		ptr = fctx->policyWorkNode;
	else
		ptr = getMemQueueFirstElementPtr(Q);
	if (!(ptr))
	{
		fctx->policyWorkNode = NULL;
		return -1;
	}
	tmp = ptr->toVisit;
	
    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	while (tmp)
	{
		if (varNthBitFromLeft(tmp, 0))
			numRelsToExtend++;
		tmp <<= 1;
	}
	tmp = ptr->toVisit;
	if (numRelsToExtend > 0)
	{
		int			k = 0;

		i = ((int) (numRelsToExtend * (rand() / (RAND_MAX + 1.0)))) + 1;
		
        Assert(i <= numRelsToExtend);
        
		j = 0;
		while ((j < i) && (j < numRelsToExtend))
		{
			if (varNthBitFromLeft(tmp, k))
			{
				j++;
			}
			k++;
		}
		k--;
		return k;
	}
	return -1;
}

/*
 * same as one at a time but with sequencing support.
 * -1 means, no relation was chosen
 */
int
policy_random_Q_one_tuple_set_at_a_time_with_sequencing(alg_fctx * fctx, memQueue * Q,
														int tsetsNum)
{
	int			i = 0;
	int			j = 0;
	int			numRelsToExtend = 0;
	RBITS		tmp = 0,
				tmp2 = 0;
	RBITS		tmpRels = 0;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	memQueueNode *ptr = NULL;

	if (fctx->policyWorkNode)
	{
		ptr = fctx->policyWorkNode;
		tmp = ptr->toVisit;
	}
	else
	{
		ptr = getMemQueueFirstElementPtr(Q);
		if (!(ptr))
			return -1;
		if (fctx->policySequence)
		{
			while (ptr != NULL)
			{
				tmpRels |= (ptr->toVisit);
				if (tmpRels == relsMask)
					break;
				ptr = getMemQueueNextElementPtr(ptr);
			}
			ptr = NULL;
			tmp = fctx->policySequence & tmpRels;
			if (!(tmp))
			{
				ptr = getMemQueueFirstElementPtr(Q);
				fctx->policySequence = relsMask;
				tmp = ptr->toVisit;
			}
		}
		else
		{
			fctx->policySequence = relsMask;
			tmp = ptr->toVisit;
		}
	}
	
    Assert(tmp);
    
	fctx->policyWorkNode = ptr;
	tmp2 = tmp;
	while (tmp)
	{
		if (varNthBitFromLeft(tmp, 0))
			numRelsToExtend++;
		tmp <<= 1;
	}
	if (numRelsToExtend > 0)
	{
		int			k = 0;

		i = ((int) (numRelsToExtend * (rand() / (RAND_MAX + 1.0)))) + 1;
		
        Assert(i <= numRelsToExtend);
        
		j = 0;
		while ((j < i) && (j < numRelsToExtend))
		{
			if (varNthBitFromLeft(tmp2, k))
			{
				j++;
			}
			k++;
		}
		k--;
		varUnSetNthBitFromLeft(fctx->policySequence, k);
		return k;
	}
	return -1;
}

/* helper macro for the policies functions */
#define update_policy_status(fctx, policy, updateCode, nodePtr)\
{ if (fctx->policyWorkNode==nodePtr)\
		fctx->policyWorkNode=NULL;}

/* Simply get a string from a datum directly.
 * must be non null value. check in advance. also check num of attr=>fnumber*/
char *
getCStrvalueFromDatum(Datum origval, TupleDesc tupdesc, int fnumber)
{
	Datum		val,
				result;
	Oid			typoid,
				foutoid;
	bool		typisvarlena;

	if (fnumber > 0)
		typoid = tupdesc->attrs[fnumber - 1]->atttypid;
	else
		typoid = (SystemAttributeDefinition(fnumber, true))->atttypid;
	getTypeOutputInfo(typoid, &foutoid, &typisvarlena);

	/*
	 * If we have a toasted datum, forcibly detoast it here to avoid memory
	 * leakage inside the type's output routine.
	 */
	if (typisvarlena)
		val = PointerGetDatum(PG_DETOAST_DATUM(origval));
	else
		val = origval;
	result = OidFunctionCall1(foutoid,
							  val);
	/* Clean up detoasted copy, if any */
	if (val != origval)
		pfree(DatumGetPointer(val));
	return DatumGetCString(result);
}

/* algorithm's extend procedure.
 * The extending operation is to take tuple sets and add to them tuples from the inputted
 * relations that the resulting set is join consistend and connected (JCC).
 * Finally stop when all the tuple sets are maximal in that respect.*/
void
Extend(alg_fctx * fctx, memQueue * Q)
{
	int			QNumElements = getMemQueueNumElements(Q);

	if (!(QNumElements))
		return;
    
    /* an estimate for the average size of a tuple set in memory. */
	int			avgTSetSize = (int) (getMemQueueSize(Q) / (QNumElements + 1)) + QNumElements;
	int			i = 0;
    
    /* The first element in Q */
	memQueueNode *nodePtr = getMemQueueFirstElementPtr(Q);
    
    /* a pointer to a queue which we need to set visited a relation to each of its tuple sets */
    memQueue   *setVisitedQ = NULL;
    
	TSet		retTSet = NULL;
	DTuple		tuple = newDTuple(fctx, fctx->resultTupleNAtt);
	
    /* a structure that deforms a tuple one attribute at a time until it is deformed completly.
     * However, we can abort in the middle */
    deformTupleIterativeState *dsTuple
	= (deformTupleIterativeState *) palloc(sizeof(deformTupleIterativeState));
	bool		endScan = false;
	bool		firstPass = true;
	bool		isNull = false;
	memQueueNode *prevNodePtr = NULL;
	TSet		tset = NULL;

	i = 0;
	bool		scanNotEmpty = false;
	int			relToCheck = -1;
	char	   *query = fctx->query;
	int			querySize = fctx->querySize;
	bool		toFreeQuery = true;
	POLICY_FUNC policy = NULL;

    /* Lets choose the inputted policy function to use, 
     * these functions all have the same interface */
	switch (fctx->extendingPolicy)
	{
		case POLICY_MAJORITY:
			if (fctx->trueIndexFalseNonIndexBasedAlgorithm)
				policy = &policy_majority_Q_one_tuple_set_at_a_time_with_sequencing;
			else
				policy = &policy_majority_Q_one_tuple_set_at_a_time;
			break;
		case POLICY_MINCOST:
			if (fctx->trueIndexFalseNonIndexBasedAlgorithm)
				policy = &policy_mincost_Q_one_tuple_set_at_a_time_with_sequencing;
			else
				policy = &policy_mincost_Q_one_tuple_set_at_a_time;
			break;
		case POLICY_RANDOM:
			if (fctx->trueIndexFalseNonIndexBasedAlgorithm)
				policy = &policy_random_Q_one_tuple_set_at_a_time_with_sequencing;
			else
				policy = &policy_random_Q_one_tuple_set_at_a_time;
			break;
		case POLICY_NAIVE:
			policy = &policy_naive_Q;
			fctx->extendingNaivePolicyNumCompleteScans = 0;
			break;
		default:
            /* This should never happen! */
			Assert(1==2);
	}

    /* the queues for the extending process */
	memQueue	needsChanging;
	memQueue	notInvolved;
	memQueue	maximalyExtended;

	_initializeMemQueue(&needsChanging, INFINIT);
	_initializeMemQueue(&notInvolved, INFINIT);
	_initializeMemQueue(&maximalyExtended, INFINIT);
    
	/* Here we work under the assumption that Q is ALREADY MERGED!!!!!!!!!!1 */
	nodePtr = getMemQueueFirstElementPtr(Q);
    /* here we find all the already maximally extended tuple sets and move them to the result */
	prevNodePtr = nodePtr;	
    while (nodePtr != NULL)
	{
		if (!(nodePtr->toVisit))
		{
			detachNodeFromMemQueueNode(prevNodePtr, nodePtr, Q);
			update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_6, nodePtr);
			addNodeToMemQueue(nodePtr, &maximalyExtended);
			nodePtr = prevNodePtr;
		}
		getMemQueueNextElementAfterRemovalPtr(prevNodePtr, nodePtr, Q);
	}
	QNumElements = getMemQueueNumElements(Q);
	fctx->extendingLastRelScanned = -1;
    
    /* this is the main iteration loop where we choose  the next relation to extend to and
     * indeed extend to it all the tuple sets that needs to extend to it */
	while ((relToCheck = policy(fctx, Q, QNumElements)) > -1)
	{
		_initializeMemQueue(&needsChanging, INFINIT);
		_initializeMemQueue(&notInvolved, INFINIT);
        
        /* The last relation we checked. */
		fctx->extendingLastRelScanned = relToCheck;
		bool		attemptIndexScan = true;

        /* before attempting to scan a relation using indices, we check it it has indices at all */
		if (fctx->noIndexScans || (!fctx->relsHasIndex[relToCheck]))
			attemptIndexScan = false;
        
		bool		firstTSet = true;
		bool		firstAtt = true;
		int			paramNum = -1;

		nodePtr = getMemQueueFirstElementPtr(Q);
        /* Here we check if we SHOULD scan using an index and formulate a proper
         * query to retrieve the needed tuples */
		if (attemptIndexScan)
		{
			int			numTSetsInQuery = 0;

			while (nodePtr)
			{
				tset = getMemQueueElement(nodePtr);
				if (varNthBitFromLeft(nodePtr->toVisit, relToCheck))
				{
					numTSetsInQuery++;
				}
				nodePtr = getMemQueueNextElementPtr(nodePtr);
			}
            
			if (!(numTSetsInQuery > 0))
				attemptIndexScan = false;
			else
			{
				if (fctx->relsPages[relToCheck] >
					(INDEX_AND_PLAN_BASE_FACTOR
					 + ((INDEX_AND_PLAN_PER_TUPLE_FACTOR_CPU_COST + INDEX_AND_PLAN_PER_TUPLE_FACTOR) * numTSetsInQuery)))
					attemptIndexScan = true;
				else
					attemptIndexScan = false;
			}
            
			if (query == NULL && attemptIndexScan)
			{
				if (fctx->querySize < 1)
				{
					int			resultAttNamesSize = 0;

					for (i = 0; i < fctx->resultTupleNAtt; i++)
						resultAttNamesSize += strlen(fctx->resultTupleAttNames[i]);
					querySize = 256 + (avgTSetSize * numTSetsInQuery)
						+ (10 + resultAttNamesSize) * fctx->relsNAtt[relToCheck];
					fctx->querySize = querySize;
				}
				query = (char *) palloc(querySize);
			}
            
            /* so we confirmed we think we should use an index. Here we formulate a proper query*/
			if (attemptIndexScan)
			{
				fctx->queryParams = getAndExpandDatumArray(fctx->queryParams
														   ,fctx->queryParamsSize, numTSetsInQuery * fctx->resultTupleNAtt);
				fctx->queryParamsTypes = getAndExpandOidArray(fctx->queryParamsTypes
															  ,fctx->queryParamsSize, numTSetsInQuery * fctx->resultTupleNAtt);
				if (fctx->queryParamsSize < numTSetsInQuery * fctx->resultTupleNAtt)
					fctx->queryParamsSize = numTSetsInQuery * fctx->resultTupleNAtt;
				Datum	   *params = fctx->queryParams;
				Oid		   *types = fctx->queryParamsTypes;
				char		paramStr[12];

				paramNum = 0;
				char	   *to = (char *) query;

				nodePtr = getMemQueueFirstElementPtr(Q);
				while (nodePtr)
				{
					tset = getMemQueueElement(nodePtr);
					if (varNthBitFromLeft(nodePtr->toVisit, relToCheck))
					{
						if (!firstTSet)
							query = addAndExpandStringBy(query, &querySize, " UNION ALL", BLCKSZ, &to
														 ,&toFreeQuery);
						firstTSet = false;
						query = addAndExpandStringBy(query, &querySize, "("
												 ,BLCKSZ, &to, &toFreeQuery);
						query = addAndExpandStringBy(query, &querySize, fctx->queries[relToCheck]
												 ,BLCKSZ, &to, &toFreeQuery);
						firstAtt = true;
						for (i = 0; i < fctx->relsNAtt[relToCheck]; i++)
						{
							if (tset->n[fctx->relsAttNumsAtTupleSet[relToCheck][i]] != 'n')
							{
								if (firstAtt)
								{
									query = addAndExpandStringBy(query, &querySize, " WHERE "
												 ,BLCKSZ, &to, &toFreeQuery);
									firstAtt = false;
								}
								else
								{
									query = addAndExpandStringBy(query, &querySize, " AND "
												 ,BLCKSZ, &to, &toFreeQuery);
								}
								query = addAndExpandStringBy(query, &querySize,
															 (char *) get_attname(fctx->relsOid[relToCheck],
																				  fctx->relsValidAttNumAtRelation[relToCheck][i])
												 ,BLCKSZ, &to, &toFreeQuery);
								params[paramNum] = tset->v[fctx->relsAttNumsAtTupleSet[relToCheck][i]];
								types[paramNum] = fctx->relsAttTypes[relToCheck][i];
								paramNum++;
								sprintf(paramStr, "%d", paramNum);
								query = addAndExpandStringBy(query, &querySize, "=$"
												 ,BLCKSZ, &to, &toFreeQuery);
								query = addAndExpandStringBy(query, &querySize, paramStr
												 ,BLCKSZ, &to, &toFreeQuery);
							}
						}
						query = addAndExpandStringBy(query, &querySize, " LIMIT 1)", BLCKSZ, &to
													 ,&toFreeQuery);
					}
					nodePtr = getMemQueueNextElementPtr(nodePtr);
				}
			}
			fctx->querySize = querySize;
			fctx->query = query;
		}
        
		_SPI_plan  *plan = NULL;
		Portal		portal = NULL;
		bool		usingSeqScan = true;
        /* here we prepare the index based query. and check if the plan includes a sequential
         * scan, in which case we should just use our already opened cursor to the relation.
         * This is done because of deficiency in the capability of the planner to accommodate
         * our complex index based query */
		if (attemptIndexScan)
		{
			if ((plan = SPI_prepare(query, paramNum, fctx->queryParamsTypes)) == NULL)
               elog(ERROR, 
                   "At %s(%d) - Could not initialize SPI structures as SPI_prepare (%s) returned NULL",
                    __FILE__, __LINE__, query);

			Plan	   *planTree;
#if (CATALOG_VERSION_NO >= 200702202)
			List *stmt_list = (List *) linitial(plan->stmt_list_list);
			Assert(list_length(plan->stmt_list_list) == 1 &&
				list_length(stmt_list) == 1 &&
				IsA((Node *) linitial(stmt_list), PlannedStmt));
			planTree = ((PlannedStmt *) linitial(stmt_list))->planTree;
#else
			ListCell   *plan_list_item = list_head(plan->ptlist);
			planTree = lfirst(plan_list_item);
#endif
			usingSeqScan = isPlanUsingSeqScan(planTree);
		}
		else
		{
			usingSeqScan = true;
		}

        
		if (usingSeqScan)
		{
			if (attemptIndexScan)
			{
				SPI_freeplan(plan);
			}
			plan = fctx->plans[relToCheck];
			portal = fctx->portals[relToCheck];
			SPI_cursor_move(portal, false, FETCH_ALL);
			SPI_freetuptable(SPI_tuptable);
		}
		else
		{
			if ((portal
				 = SPI_cursor_open(NULL, plan, fctx->queryParams, NULL, true)) == NULL)
               elog(ERROR, 
                "At %s(%d) - Could not initialize SPI structures as SPI_cursor_open (%s) returned NULL", 
                 __FILE__, __LINE__, query);
		}

        /* Lets remove a tuple from either our usual portal or specially opened index
         * based portal */
		SPI_cursor_fetch(portal, true, 1);
        
		retTSet = NULL;
		endScan = false;
		firstPass = true;
		prevNodePtr = NULL;
		scanNotEmpty = false;

        /* This loop handles extending of all the tuple sets to the relation we are currently
         * attempting to extend to. */
		while ((SPI_processed > 0) && (!endScan))
		{
			tuple->t = SPI_tuptable->vals[0];
			tuple->deformed = false;
			dsTuple->firstIteration = true;
			scanNotEmpty = true;
			isNull = false;
			if (fctx->extendingPolicy == POLICY_NAIVE)
				endScan = false;
			else
				endScan = true;
            /* basically we move tuple sets based on their charecteristics of their potential
             * to be extended (need extend, already extend, cannot extend...) 
             * The first pass does the initial redirections and attempt to extend to the first tuple */
			if (firstPass)
			{
				detachHeadNodeFromMemQueue(nodePtr, Q);
				while (nodePtr)
				{
					if (varNthBitFromLeft(nodePtr->toVisit, relToCheck))
					{
						tset = getMemQueueElement(nodePtr);
						endScan = false;
						retTSet = JCCTSetAndTupleAndReturnTSet(fctx, tset, tuple
								,SPI_tuptable->tupdesc, relToCheck, dsTuple);
						if (retTSet != NULL)
						{
							varSetNthBitFromLeft(nodePtr->relsOfTuples, relToCheck);
							varUnSetNthBitFromLeft(nodePtr->toVisit, relToCheck);
							nodePtr->toVisit |= ((fctx->bits_scheme_graph[relToCheck])
												 & (~(nodePtr->relsOfTuples)) & (~(nodePtr->visited)));
							if (nodePtr->toVisit)
							{
								mergeOutOfMemQueueIntoNode(fctx, &needsChanging, nodePtr, false, true);
								mergeOutOfMemQueueIntoNode(fctx, Q, nodePtr, false, true);
								if (nodePtr->toVisit)
								{
									update_policy_status(fctx, policy, POLICY_CODE_NOT_INVOLVED_1, nodePtr);
									addNodeToMemQueue(nodePtr, &notInvolved);
								}
								else
								{
									update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_0,
														 nodePtr);
									addNodeToMemQueue(nodePtr, &maximalyExtended);
								}
							}
							else
							{
								update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_1,
													 nodePtr);
								addNodeToMemQueue(nodePtr, &maximalyExtended);
							}
						}
						else
							addNodeToMemQueue(nodePtr, &needsChanging);
					}
					else if (!(nodePtr->toVisit))
					{
						update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_2, nodePtr);
						addNodeToMemQueue(nodePtr, &maximalyExtended);
					}
					else
					{
						update_policy_status(fctx, policy, POLICY_CODE_NOT_INVOLVED_2, nodePtr);
						addNodeToMemQueue(nodePtr, &notInvolved);
					}
					detachHeadNodeFromMemQueue(nodePtr, Q);
				}
				firstPass = false;
			}
			else
			{
				nodePtr = getMemQueueFirstElementPtr(&needsChanging);
				prevNodePtr = nodePtr;
				while (nodePtr)
				{
					tset = getMemQueueElement(nodePtr);
					endScan = false;
					retTSet = JCCTSetAndTupleAndReturnTSet(fctx, tset, tuple
								,SPI_tuptable->tupdesc, relToCheck, dsTuple);
					if (retTSet != NULL)
					{
						varSetNthBitFromLeft(nodePtr->relsOfTuples, relToCheck);
						varUnSetNthBitFromLeft(nodePtr->toVisit, relToCheck);
						nodePtr->toVisit |= ((fctx->bits_scheme_graph[relToCheck])
						& (~(nodePtr->relsOfTuples)) & (~(nodePtr->visited)));
						detachNodeFromMemQueueNode(prevNodePtr, nodePtr, &needsChanging);
						if (nodePtr->toVisit)
						{
							mergeOutOfMemQueueIntoNode(fctx, &needsChanging, nodePtr, false, true);
							if (nodePtr->toVisit)
							{
								update_policy_status(fctx, policy, POLICY_CODE_NOT_INVOLVED_3, nodePtr);
								addNodeToMemQueue(nodePtr, &notInvolved);
							}
							else
							{
								update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_3,
													 nodePtr);
								addNodeToMemQueue(nodePtr, &maximalyExtended);
							}
						}
						else
						{
							update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_4,
												 nodePtr);
							addNodeToMemQueue(nodePtr, &maximalyExtended);
						}
						nodePtr = prevNodePtr;
					}
					getMemQueueNextElementAfterRemovalPtr(prevNodePtr, nodePtr, &needsChanging);
				}
			}
			SPI_freetuptable(SPI_tuptable);
			SPI_cursor_fetch(portal, true, 1);
		}
		SPI_freetuptable(SPI_tuptable);

        /* we finished here extending to the specific relation. 
         * we now set all the non extended tuple sets to being visited. */
		if (!scanNotEmpty)
			setVisitedQ = Q;
		else
			setVisitedQ = &needsChanging;
		nodePtr = getMemQueueFirstElementPtr(setVisitedQ);
		prevNodePtr = nodePtr;
		while (nodePtr != NULL)
		{
			if (varNthBitFromLeft(nodePtr->toVisit, relToCheck))
			{
				varUnSetNthBitFromLeft(nodePtr->toVisit, relToCheck);
				varSetNthBitFromLeft(nodePtr->visited, relToCheck);
			}
			if (!(nodePtr->toVisit))
			{
				detachNodeFromMemQueueNode(prevNodePtr, nodePtr, setVisitedQ);
				update_policy_status(fctx, policy, POLICY_CODE_MAXIMALLY_EXTENDED_5, nodePtr);
				addNodeToMemQueue(nodePtr, &maximalyExtended);
				nodePtr = prevNodePtr;
			}
			getMemQueueNextElementAfterRemovalPtr(prevNodePtr, nodePtr, setVisitedQ);
		}

        /* finally lets return all the tuple sets we redirected based on their extending potential
         * to other queues, to Q to start the whole process again. */
		if (scanNotEmpty)
			recombineQueues(Q, 2, &needsChanging, &notInvolved);
		QNumElements = getMemQueueNumElements(Q);
		if (!usingSeqScan)
		{
			SPI_cursor_close(portal);
			SPI_freeplan(plan);
		}
	}
	pfree(tuple);
	pfree(dsTuple);
	Assert(isMemQueueEmpty(Q));
	Assert(isMemQueueEmpty(&notInvolved));
	Assert(isMemQueueEmpty(&needsChanging));

    /* finally we move all the maximally extended tuple sets back into Q to return from the function */
	recombineQueues(Q, 1, &maximalyExtended);
}

/* 
 * This procedure updates the relations we need to scan for generating 
 * alternative tuple sets (algorithm FD terminology) which greatly reduces
 * the time of the algorithm.
 * Note, that here we changed the meaning of ptr->toVisit to mean that
 * we have to generate alternative tuple set with a relation in toVisit.*/
void
updateOnDemandRelationsToScan(alg_fctx * fctx, memQueue * Q)
{
	RBITS		relationsToScan = 0;
	int			i = 0,
				j = 0,
				k = 0;
	memQueueNode *ptr = getMemQueueFirstElementPtr(Q);
	int			numAtt = NUMATT;
	int			numOfTuples;
	int			numOfTuplesInIntersection = 0;
	int			tsetAtt;
	int			tsetAttributes[numAtt];
	RBITS		RNeighboursIntersectTSetRelations = 0;
	RBITS		relsOfTuples;
	RBITS		relsMask = (~0x0);

	relsMask >>= fctx->numRels;
	relsMask = ~relsMask;
	TSet		tset = NULL;
	bool		addRelation = false;

    /* we pass over all the tuple sets we wish to generate alternative tuple sets
     * for and update the list of relations we will need to scan if we suspect it
     * is possible we will need to scan it for some tuple sets.
     * The algorithm itself is in the FD algorithm docs. */
	while (ptr)
	{
		tset = ptr->tset;
		ptr->toVisit = 0;
		relsOfTuples = ptr->relsOfTuples;
		numOfTuples = numOfTuplesInTSet(fctx, tset->tids);
		if (numOfTuples > 1)
		{
			relationsToScan |= relsOfTuples;
			ptr->toVisit |= relsOfTuples;
			for (i = 0; i < fctx->numRels; i++)
			{
				for (k = 0; k < numAtt; k++)
					tsetAttributes[k] = 0;
				RNeighboursIntersectTSetRelations = (fctx->bits_scheme_graph[i] & relsOfTuples);
				if (!(RNeighboursIntersectTSetRelations))
					continue;
				j = 0;
				numOfTuplesInIntersection = 0;
				while (RNeighboursIntersectTSetRelations)
				{
					if (varNthBitFromLeft(RNeighboursIntersectTSetRelations, 0))
					{
						numOfTuplesInIntersection++;
						for (k = 0; k < fctx->relsNAtt[j]; k++)
						{
							tsetAttributes[fctx->relsAttNumsAtTupleSet[j][k]]++;
						}
					}
					j++;
					RNeighboursIntersectTSetRelations <<= 1;
				}
				addRelation = false;
				for (k = 0; k < fctx->relsNAtt[i]; k++)
				{
					tsetAtt = fctx->relsAttNumsAtTupleSet[i][k];
					if (tsetAttributes[tsetAtt] == numOfTuplesInIntersection)
					{
						if (tsetAttIsNull(tset, tsetAtt))
						{
							addRelation = false;
							break;
						}
					}
					else if (tsetAttributes[tsetAtt] > 0)
					{
						addRelation = true;
					}
				}
				if (addRelation)
				{
					varSetNthBitFromLeft(ptr->toVisit, i);
					varSetNthBitFromLeft(relationsToScan, i);
				}
			}
		}
		else if (!(relsOfTuples))
		{
			ptr->toVisit = relsMask;
			relationsToScan = relsMask;
		}
		ptr = getMemQueueNextElementPtr(ptr);
	}
	fctx->relationsToScan = relationsToScan;
}


/* here we remove one page from preComplete into Q */
void
onDemandPreCompleteRemovePage(alg_fctx * fctx, FuncCallContext *funcctx, memQueue * Q)
{
	clearMemQueue(fctx, Q, true);
	bool		should_free;
	HeapTuple	tuptset;
	int			readSize = 0;

	while (SQNotEmpty(&fctx->precomplete))
	{
		tuptset = SQGetNextElement(fctx, funcctx, &fctx->precomplete, &should_free, true);
		readSize += tupleSize(tuptset);
		if (readSize > (BLCKSZ))
		{
			SQSetMemOverflowTuple(&fctx->precomplete, tuptset);
			break;
		}
		else
		{
			addTupleSetToMemQueue(fctx, Q, newTSet___(fctx, tuptset));
		}
	}
	updateOnDemandRelationsToScan(fctx, Q);
}


/* Here we take a relation to scan from the list of relations to scan for
 * the on demand procedure */
int
getRandomNextRelationToScan(alg_fctx * fctx, int lastScannedRelation)
{
	RBITS		relationsToScan = fctx->relationsToScan;
	RBITS		tmp = fctx->relationsToScan;
	int			numRelsToScan = 0;

	while (tmp)
	{
		if (varNthBitFromLeft(tmp, 0))
			numRelsToScan++;
		tmp <<= 1;
	}
	if (numRelsToScan > 0)
	{
		int			j = 0;
		int			k = 0;
		int			i = ((int) (numRelsToScan * (rand() / (RAND_MAX + 1.0)))) + 1;

		Assert(i <= numRelsToScan);

		while ((j < i) && (j < numRelsToScan))
		{
			if (varNthBitFromLeft(relationsToScan, k))
			{
				j++;
			}
			k++;
		}
		k--;
		varUnSetNthBitFromLeft(fctx->relationsToScan, k);
		return k;
	}
	return -1;
}

/* Here we retrieve one tuple at a time from the relation we are currently scanning
 * and if we reach the end of the relation we cycle to the next relation in the list
 * automatically. The name "..NextPage.." is left to accommodate the original terminology
 * of the FD algorithm, however, the page is indeed read in the backend and there is no
 * actual need to load a whole page again here. */
void
onDemandRetrieveNextPageFromR(alg_fctx * fctx, FuncCallContext *funcctx)
{
	OnDemandState *S = fctx->onDemandState;

	if (S->lastTuptable)
	{
		SPI_freetuptable(S->lastTuptable);
		S->lastTuptable = NULL;
	}
	if (S->P->t == NULL)
	{
		S->cur_rel = getRandomNextRelationToScan(fctx, -1);
		if (S->cur_rel >= 0)
		{
			SPI_cursor_move(S->portals[S->cur_rel], false, FETCH_ALL);
			SPI_freetuptable(SPI_tuptable);
		}
	}
	while (S->cur_rel >= 0)
	{
		SPI_cursor_fetch(S->portals[S->cur_rel], true, 1);
		if (SPI_processed > 0)
		{
			S->P->t = SPI_tuptable->vals[0];
			S->P->deformed = false;
			S->Pdesc = SPI_tuptable->tupdesc;
			S->PrelID = S->cur_rel;
			S->lastTuptable = SPI_tuptable;
			return;
		}
		SPI_freetuptable(SPI_tuptable);
		if ((S->cur_rel = getRandomNextRelationToScan(fctx, S->cur_rel)) < 0)
			break;
		SPI_cursor_move(S->portals[S->cur_rel], false, FETCH_ALL);
		SPI_freetuptable(SPI_tuptable);
	}
	S->P->t = NULL;
	return;
}

/* This procedure is the On demand procedure which is a FD algorithm specific
 * terminology for generating alternative tuple sets on demand, i.e., retrieving
 * enough alternative tuple sets to fit into allocated memory. The implementation
 * here is very close to the implementation suggested in the FD algorithm docs.
 * The inputs are also documented in the FD algorithm docs.*/
void
OnDemandGetAlternativeTupleSets(alg_fctx * fctx, FuncCallContext *funcctx,
		  memQueue * out, bool trueUseMemorySizeFalseUseNumTupleSets, int SZ,
								int numTupleSetsToReturn, bool addToComplete)
{
	OnDemandState *S = fctx->onDemandState;
	TSet		memQTSet = NULL;
	TSet		alternativeTSet = NULL;

	if (isMemQueueEmpty(S->Q))
	{
		fctx->relationsToScan = 0;
        /* here we read a page from precomplete and a tuple from the relations to scan */
		while (!(fctx->relationsToScan))
		{
			onDemandPreCompleteRemovePage(fctx, funcctx, S->Q);
			if (addToComplete)
			{
				memQueueNode *ptr = getMemQueueFirstElementPtr(S->Q);
				TSet		tset = NULL;
				int			i;
				char		nulls[NUMATT];

				for (i = FIRSTATT; i < NUMATT; i++)
					nulls[i] = 'n';
				nulls[LABELS] = ' ';
				HeapTuple	tempTuple;

				while (ptr != NULL)
				{
					tset = getMemQueueElement(ptr);
					if (numOfTuplesInTSet(fctx, tset->tids) > 0)
					{
						tempTuple = clearValuesFromTSet(fctx, tset, nulls);
						SQAddElement(fctx, funcctx, &fctx->complete, tempTuple);
						heap_freetuple(tempTuple);
					}
					ptr = getMemQueueNextElementPtr(ptr);
				}
			}
			if (isMemQueueEmpty(S->Q))
				return;
		}
		S->P->t = NULL;
		S->p = 1;
		S->q = 0;
		S->qPtr = getMemQueueFirstElementPtr(S->Q);
		onDemandRetrieveNextPageFromR(fctx, funcctx);
	}
	while (1)
	{
		if (trueUseMemorySizeFalseUseNumTupleSets && numPagesInMemQueue(out) >= SZ)
			return;
		else if (!trueUseMemorySizeFalseUseNumTupleSets
				 && numTupleSetsToReturn <= getMemQueueNumElements(out))
			return;
		if (S->q > 0)
			S->qPtr = getMemQueueNextElementPtr(S->qPtr);
		S->q++;
		if (S->q > getMemQueueNumElements(S->Q))
		{
			S->p++;
			S->q = 1;
			S->qPtr = getMemQueueFirstElementPtr(S->Q);
		}
		if (S->p > 1)
		{
			S->p = 1;
			onDemandRetrieveNextPageFromR(fctx, funcctx);
		}
		if (S->P->t == NULL)
		{
			fctx->relationsToScan = 0;
			while (!(fctx->relationsToScan))
			{
				onDemandPreCompleteRemovePage(fctx, funcctx, S->Q);
				if (addToComplete)
				{
					memQueueNode *ptr = getMemQueueFirstElementPtr(S->Q);
					TSet		tset = NULL;
					int			i;
					char		nulls[NUMATT];

					for (i = FIRSTATT; i < NUMATT; i++)
						nulls[i] = 'n';
					nulls[LABELS] = ' ';
					HeapTuple	tempTuple;

					while (ptr != NULL)
					{
						tset = getMemQueueElement(ptr);
						if (numOfTuplesInTSet(fctx, tset->tids) > 0)
						{
							tempTuple = clearValuesFromTSet(fctx, tset, nulls);
							SQAddElement(fctx, funcctx, &fctx->complete, tempTuple);
							heap_freetuple(tempTuple);
						}
						ptr = getMemQueueNextElementPtr(ptr);
					}
				}
				if (isMemQueueEmpty(S->Q))
					return;
			}
			S->p = 1;
			S->q = 1;
			S->qPtr = getMemQueueFirstElementPtr(S->Q);
			onDemandRetrieveNextPageFromR(fctx, funcctx);
		}
        /* here we generate the alternative tuple set with a tuple set from precomplete
         * and a tuple from the relation */
		memQTSet = getMemQueueElement(S->qPtr);
		if ((S->qPtr->relsOfTuples) && (varNthBitFromLeft(S->qPtr->toVisit, S->PrelID)))
		{
			alternativeTSet = generateAlternativeTupleSet(fctx, memQTSet, S->qPtr->relsOfTuples,
												  S->P, S->Pdesc, S->PrelID);
		}
        /* this handles the singleton tuple sets that are generated by taking the empty tuple
         * set and generating alternative tuple sets to it with tuples from the relations which
         * trivially results in the singleton tuple set that holds the tuple from the relation */
		else if (!(S->qPtr->relsOfTuples))
		{
			makeSureDeformedTuple(fctx, S->P, S->Pdesc, S->cur_rel);
			alternativeTSet = create_tuple_set(fctx, S->P, S->Pdesc, S->PrelID);
		}
		else
			alternativeTSet = NULL;
        
        /* finally if we indeed generated the alternative tuple set we output it */
		if (alternativeTSet != NULL)
		{
			addTupleSetToMemQueue(fctx, out, alternativeTSet);
		}
	}
}

/* simple floor function */
int
algFloor(float value)
{
	return (int) value;
}

/* simple ceiling function */
int
algCeiling(float value)
{
	if (value > (int) value)
		return (((int) value) + 1);
	else
		return (int) value;
}

/* simple max function */
double
max(double a, double b)
{
	if (a > b)
		return a;
	else
		return b;
}

/*simple min function */
double
min(double a, double b)
{
	if (a < b)
		return a;
	else
		return b;
}

/* getting the upper bound calculated by the adaptive algorithm */
int
getAdaptiveUpperBound(alg_fctx * fctx)
{
	if (fctx->dynamic_extend_cap->stage > 0)
		return fctx->dynamic_extend_cap->nextUpperBound;
	else
		return ADAPTIVE_EXTEND_UPPER_BOUND_MAX;
}

/* the core of the adaptive algorithm of an upper bound for the number of
 * tuple sets to be extended in the extending process at once.
 * This is not documented in the FD algorithm docs but it is an adhoc
 * algorithm created for efficiency reasons by using trial and errors.
 * It is a mathematical formulation which is straight forward from the code. */
void
updateAdaptiveScore(alg_fctx * fctx, double time)
{
	adaptive_extend_upperbound *ad = fctx->dynamic_extend_cap;
	double		currentScore = ((ad->actualInputedExtendTuples ^ 2) / time);

	if (ad->stage == 0)
	{
		ad->previousScore = currentScore;
		if ((2 * ad->maxInputedExtendTuples) <= ADAPTIVE_EXTEND_UPPER_BOUND_MAX)
			ad->nextUpperBound = 2 * ad->maxInputedExtendTuples;
		else
			ad->nextUpperBound = ADAPTIVE_EXTEND_UPPER_BOUND_MAX;
		ad->stage = 1;
		return;
	}
	if (ad->stage == 1)
	{
		if (currentScore > ad->previousScore)
		{
			if (ad->actualInputedExtendTuples > ad->nextUpperBound)
			{
				ad->nextUpperBound = ad->actualInputedExtendTuples;
			}
		}
		else
		{
			ad->stage = 2;
		}
	}
	if (ad->stage == 2)
	{
		int			currentUpperBound = ad->nextUpperBound;
		double		halfNormalizedDiffScore
		= (max(currentScore, ad->previousScore)
		   - min(currentScore, ad->previousScore)) / (2 * max(currentScore, ad->previousScore));

		if (currentScore > ad->previousScore)
			ad->nextUpperBound = ad->nextUpperBound + ADAPTIVE_EXTEND_UPPER_BOUND_FIXED_PENALTY_INC
				+ (int) (halfNormalizedDiffScore * ad->nextUpperBound);
		else
			ad->nextUpperBound = ad->nextUpperBound -
				(ADAPTIVE_EXTEND_UPPER_BOUND_FIXED_PENALTY_DEC + (int) (halfNormalizedDiffScore * ad->nextUpperBound));
		if (ad->nextUpperBound < 100)
			ad->nextUpperBound = currentUpperBound;
	}
	ad->previousScore = currentScore;
}
