/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#include "postgres.h"
#include "funcapi.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"
#include "executor/spi_priv.h"
#include  <sys/time.h>
#include "utils/lsyscache.h"
#include "utils/typcache.h"
#include "utils/elog.h"
#include "utils/memutils.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif   /* */

#include "algstructs.h"
#include "tset.h"
#include "tsetfuncs.h"
#include "queues.h"
#include "queuesfuncs.h"
#include "algutils.h"

#define DEFAULT_MB 100
#define PRINT 1

enum odmbfd_algorithm_state
{
	ALGORITHM_START = 0,
    /* Load all the tuples from the relations into the virtual PreComplete */
	MAIN_PRELOADING,
	MAIN_WHILE_PRECOMPLETE_NOTEMPTY,
	MAIN_WHILE_Q_NOTEMPTY,
    /* Output the result tuple sets */
	MAIN_PRINT,
	ALGORITHM_END
};

void
initialize_odmbfd_structures(PG_FUNCTION_ARGS, alg_fctx * fctx, FuncCallContext *funcctx)
{
	/* initialize tuple stores. */
	SQInitialize(fctx, funcctx, &fctx->complete);
	SQInitialize(fctx, funcctx, &fctx->precomplete);
    
	/* additional, per algorithm arguments (for future algorithms) */
	if (fctx->nargs >= 6)
	{
		/* MEMORY MUST BE ALLOCATED IN PAGES! */
		fctx->MB = PG_GETARG_INT32(5);
		if (fctx->MB < 2)
            ereport(ERROR,
			 (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
			  errmsg("problem with MB parameter!"),
			  errhint("The memory allocation must be greater than 1, prefferably at least as the number of relations inputted and bigger is faster (commonly at least 100)!")));
	}
	else
		fctx->MB = DEFAULT_MB;

	_initializeMemQueue(&fctx->Q, (fctx->MB * BLCKSZ));

	/*
	 * potentially 1 page (= BLCKSZ) and MB pages that could be free (however,
	 * the algorithm will determine how much to use)
	 */
	_initializeMemQueue(&fctx->E, ((1 + fctx->MB) * BLCKSZ));
    
	/* persistant state for the On Demand procedure. */
	fctx->onDemandState = (OnDemandState *) palloc(sizeof(OnDemandState));
	OnDemandState *S = fctx->onDemandState;

	S->Q = palloc(sizeof(memQueue));
	S->lastTuptable = NULL;
	S->portals = (Portal *) palloc(fctx->numRels * sizeof(Portal));
	S->portals[0] = NULL;
	{
		int			i;
		int			maxRelNatt = 0;

		for (i = 0; i < fctx->numRels; i++)
			if (maxRelNatt < fctx->relsNAtt[i])
				maxRelNatt = fctx->relsNAtt[i];
		S->P = newDTuple(fctx, maxRelNatt);
	}
	int			i = 0;
	for (i = 0; i < fctx->numRels; i++)
	{
		if ((S->portals[i]
		  = SPI_cursor_open(NULL, fctx->plans[i], NULL, NULL, true)) == NULL)

			/*
			 * note that read_only in cursor_open must be true otherwise ctid
			 * can change.
			 */
            elog(ERROR, 
                 "At %s(%d) - Could not initialize SPI structures as SPI_cursor_open (PLAN: %d) returned NULL", 
                  __FILE__, __LINE__, i);
	}
	_initializeMemQueue(S->Q, BLCKSZ);
	
	/* Here we calculate the maximal size of a tuple for each relation. */
    {
		int			i = 0;

		for (i = 0; i < fctx->numRels; i++)
		{
			SPI_cursor_move(fctx->portals[i], false, FETCH_ALL);
			SPI_cursor_fetch(fctx->portals[i], true, 1);
			while (SPI_processed > 0)
			{
				updateMaxTupleSize(fctx, i, SPI_tuptable->vals[0]);
				SPI_freetuptable(SPI_tuptable);
				SPI_cursor_fetch(fctx->portals[i], true, 1);
			}
			SPI_freetuptable(SPI_tuptable);
			SPI_cursor_move(fctx->portals[i], false, FETCH_ALL);
		}
		updateResultTupleMaxSize(fctx);
		fctx->assumingResultTupleMaxSize = false;
	}
}

/*
 * Here we free the most important to free resources.
 */
void
clean_odmbfd_structures_on_exit(PG_FUNCTION_ARGS, alg_fctx * fctx
								,FuncCallContext *funcctx)
{
	SQFinish(fctx, funcctx, &fctx->precomplete);
	SPI_finish();
	MemoryContextDelete(fctx->per_query_ctx);
}

PG_FUNCTION_INFO_V1(odmbfd);
Datum
odmbfd(PG_FUNCTION_ARGS)
{
	FuncCallContext *funcctx;
	alg_fctx   *fctx = NULL;

	/* stuff done only on the first call of the function */
	if (SRF_IS_FIRSTCALL())
	{
		MemoryContext oldcontext;

		/* create a function context for cross-call persistence */
		funcctx = SRF_FIRSTCALL_INIT();
        
		/* switch to memory context appropriate for multiple function calls */
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        
		fctx = (alg_fctx *) palloc(sizeof(alg_fctx));
		
        fctx->trueIndexFalseNonIndexBasedAlgorithm = false;
        
        /* Initialize general to all future algorithms structures */
		initialize_general_structures(fcinfo, fctx, funcctx);

        Assert(fctx != NULL);
		
        funcctx->tuple_desc = BlessTupleDesc(fctx->resultTupleDesc);
        
        /* initialize general SPI structures */
		initialize_SPI_structures(fcinfo, fctx, !fctx->noDuplicates, funcctx);
		
        /* initialize this algorithm specific structures */
        initialize_odmbfd_structures(fcinfo, fctx, funcctx);
		
        fctx->algorithm_state = ALGORITHM_START;
		
        funcctx->user_fctx = fctx;
		
        MemoryContextSwitchTo(oldcontext);
	}
    
    
	/* stuff done on every call of the function */
	funcctx = SRF_PERCALL_SETUP();
	fctx = funcctx->user_fctx;
	/* Seems that if i take this off, all kind of leaks. Better to leave this. */
	MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

    
	/* Algorithm starts here. */
	while (fctx->algorithm_state != ALGORITHM_END)
	{
		switch (fctx->algorithm_state)
		{
			case ALGORITHM_START:
				clearMemQueue(fctx, &fctx->Q, true);
				fctx->algorithm_state = MAIN_PRELOADING;
				break;
			case MAIN_PRELOADING:
				{
                  /* here we load an empty set that its union with each tuple of the relation is
                   * the singleton tuple of the relation itself */
					TSet		emptyTSet = newEmptyTSet(fctx);

					SQAddElement(fctx, funcctx, &fctx->precomplete, emptyTSet->t);
				}
				fctx->algorithm_state = MAIN_WHILE_PRECOMPLETE_NOTEMPTY;
				break;
			case MAIN_WHILE_PRECOMPLETE_NOTEMPTY:
				/*
				 * the ondemand->Q check is instead of the uncommited read in
				 * ondemand function in the thesis. more efficient,
				 * programming wise.
				 */
				if (SQNotEmpty(&fctx->precomplete) || !isMemQueueEmpty(fctx->onDemandState->Q))
				{
					OnDemandGetAlternativeTupleSets(fctx, funcctx, &fctx->Q, true, fctx->MB, -1, true);
					if (!isMemQueueEmpty(&fctx->Q))
					{
						HeapTuple	tuptset = NULL;
						bool		should_free = false;
						bool		isnull;
						TSet		newtset = newTSet____(fctx);

						/* Q.RemoveSumsumptionsAgainstComplete */
						SQRewindToFirstElement(fctx, funcctx, &fctx->complete);
						while (!SQEof(&fctx->complete))
						{
							if (isMemQueueEmpty(&fctx->Q))
								break;
							tuptset = SQGetNextElement(fctx, funcctx, &fctx->complete, &should_free, false);
							setTSet(fctx, newtset, tuptset, isnull);
							removeLeftMemQueueNodesSubsumedInRightTSet(fctx, &fctx->Q, newtset);
							if (should_free)
								heap_freetuple(tuptset);
						}
						/* Q.RemoveSumsumptionsAgainstPreComplete */
						SQMarkPosition(fctx, funcctx, &fctx->precomplete);
						while (!SQEof(&fctx->precomplete))
						{
							if (isMemQueueEmpty(&fctx->Q))
								break;
							tuptset = SQGetNextElement(fctx, funcctx, &fctx->precomplete, &should_free, false);
							setTSet(fctx, newtset, tuptset, isnull);
							removeLeftMemQueueNodesSubsumedInRightTSet(fctx, &fctx->Q, newtset);
							if (should_free)
								heap_freetuple(tuptset);
						}
						SQRestorePosition(fctx, funcctx, &fctx->precomplete);
						pfree(newtset);
					}
					fctx->algorithm_state = MAIN_WHILE_Q_NOTEMPTY;
				}
				else
					fctx->algorithm_state = ALGORITHM_END;
				break;
			case MAIN_WHILE_Q_NOTEMPTY:
				clearMemQueue(fctx, &fctx->E, true);
				if (!isMemQueueEmpty(&fctx->Q))
				{
					/*
					 * New addition: adaptive upper bound on the number of
					 * tuples being extended at one time. Because of CPU
					 * constraits and transactional constraints, changing this
					 * value might be worth it.
					 */
					struct timeval startTime;

					gettimeofday(&startTime, NULL);
					int			freePagesInQ = fctx->MB - floorNumPagesInMemQueue(&fctx->Q);

					if (freePagesInQ < 0)
						freePagesInQ = 0;
					int			numTupleSetsToGet
					= fctx->m_NumResultTuplesInOnePage * (freePagesInQ + 1);

					if (fctx->dynamic_extend_cap->maxInputedExtendTuples < numTupleSetsToGet)
						fctx->dynamic_extend_cap->maxInputedExtendTuples = numTupleSetsToGet;
					if (getAdaptiveUpperBound(fctx) < numTupleSetsToGet)
						numTupleSetsToGet = getAdaptiveUpperBound(fctx);
					fctx->dynamic_extend_cap->actualInputedExtendTuples = numTupleSetsToGet;
					memQueueNode *nodePtr;

					while (!isMemQueueEmpty(&fctx->Q) && (getMemQueueNumElements(&fctx->E) < numTupleSetsToGet))
					{
						detachHeadNodeFromMemQueue(nodePtr, &fctx->Q);
						mergeOutOfMemQueueIntoNode(fctx, &fctx->E, nodePtr, true, true);
						addNodeToMemQueue(nodePtr, &fctx->E);
					}
                    
					Extend(fctx, &fctx->E);
                    
					removeLeftMemQueueNodesSubsumedInRightMemQueueNodes(fctx, &fctx->Q, &fctx->E);
					/* add E to PRECOMPLETE */
					{
						memQueueNode *ptr = getMemQueueFirstElementPtr(&fctx->E);
						TSet		tset = NULL;

						while (ptr != NULL)
						{
							tset = getMemQueueElement(ptr);
							tset->t = heap_formtuple(fctx->tupleSetDesc, tset->v, tset->n);
							SQAddElement(fctx, funcctx, &fctx->precomplete, tset->t);
							ptr = getMemQueueNextElementPtr(ptr);
						}
					}
					{
						struct timeval tim;

						gettimeofday(&tim, NULL);
						double		length = tim.tv_usec - startTime.tv_usec;

						if (length > 0)
							updateAdaptiveScore(fctx, length);
						else
							updateAdaptiveScore(fctx, 10000);
					}
					if (PRINT)
						fctx->algorithm_state = MAIN_PRINT;
				}
				else
				{
					fctx->algorithm_state = MAIN_WHILE_PRECOMPLETE_NOTEMPTY;
				}
				break;
			case MAIN_PRINT:
				if (fctx->printedTupleAddress != NULL)
				{
					pfree(fctx->printedTupleAddress);
					fctx->printedTupleAddress = NULL;
				}
				if (fctx->startedPrintingQueue == false)
				{
					fctx->printedTSet = fctx->E.head;
					fctx->startedPrintingQueue = true;
				}
				else
					fctx->printedTSet = (memQueueNode *) fctx->printedTSet->next;
				if (fctx->printedTSet != NULL)
				{
					fctx->printedTupleAddress
						= heap_formtuple(fctx->resultTupleDesc, &fctx->printedTSet->tset->v[FIRSTATT],
									  &fctx->printedTSet->tset->n[FIRSTATT]);
					SRF_RETURN_NEXT(funcctx, HeapTupleGetDatum(fctx->printedTupleAddress));
				}
				else
				{
					fctx->startedPrintingQueue = false;
					fctx->algorithm_state = MAIN_WHILE_Q_NOTEMPTY;
				}
				break;
			case ALGORITHM_END:
				break;
            default:
              /* This can't happen! */
              Assert(1==2);
		}
	}

	/* Algorithm ends here. */
	clean_odmbfd_structures_on_exit(fcinfo, fctx, funcctx);
	SRF_RETURN_DONE(funcctx);
}
