/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#include "queues.h"
#include "queuesfuncs.h"
#include "tset.h"
#include "tsetfuncs.h"
#include "algutils.h"
#include "utils/tuplestore.h"
#include "utils/memutils.h"

/* initialize a memory queue */
void
_initializeMemQueue(memQueue * Q, int allocatedBytes)
{
	Q->head = NULL;
	Q->tail = NULL;
	Q->numElements = 0;
	Q->sizeInBytes = 0;
	Q->allocatedBytes = allocatedBytes;
}

/* remove duplicate tuple sets from the memory queue */
void
removeDuplicatesMemQueueWithSelf(alg_fctx * fctx, memQueue * Q)
{
	if (Q->numElements < 2)
		return;
	memQueueNode *lPtr = Q->head;
	memQueueNode *rPtr = Q->head;
	memQueueNode *prevRPtr = Q->head;
	RBITS		lrels;

	while (lPtr != NULL)
	{
		if (Q->numElements < 2)
			return;
		lrels = lPtr->relsOfTuples;
		rPtr = (memQueueNode *) lPtr->next;
		prevRPtr = lPtr;
		while (rPtr != NULL)
		{
			if (lrels == rPtr->relsOfTuples)
				if (areTSetsEqual(fctx, lPtr->tset, rPtr->tset))
				{
					detachNodeFromMemQueueNode(prevRPtr, rPtr, Q);
					freeTSet(fctx, rPtr->tset);
					pfree(rPtr);
					rPtr = prevRPtr;
				}
			getMemQueueNextElementAfterRemovalPtr(prevRPtr, rPtr, Q);
		}
		lPtr = (memQueueNode *) lPtr->next;
	}
}

/* remove subsumptions from a memory queue with itself.
 * Subsumptions are a containtment of a tuple set in another tuple set. */
void
removeSubsumptionsMemQueueWithSelf(alg_fctx * fctx, memQueue * Q)
{
	if (Q->numElements < 2)
		return;
	memQueueNode *lPtr = Q->head;
	memQueueNode *rPtr = Q->head;
	memQueueNode *prevRPtr = Q->head;

	while (lPtr != NULL)
	{
		if (Q->numElements < 2)
			return;
		rPtr = Q->head;
		prevRPtr = Q->head;
		while (rPtr != NULL)
		{
			if (lPtr != rPtr)
			{
				if (isLeftSubsumedRightTset(fctx, lPtr->tset, rPtr->tset))
			{
					Q->sizeInBytes -= tupleSize(rPtr->tset->t);
					if ((memQueueNode *) lPtr->next == (memQueueNode *) rPtr)
						lPtr->next = rPtr->next;
					if (Q->head == rPtr)
						Q->head = (memQueueNode *) rPtr->next;
					if (rPtr == Q->tail)
						Q->tail = prevRPtr;
					prevRPtr->next = rPtr->next;
					freeTSet(fctx, rPtr->tset);
					pfree(rPtr);
					Q->numElements--;
					break;
				}
			}
			prevRPtr = rPtr;
			rPtr = (memQueueNode *) rPtr->next;
		}
		lPtr = (memQueueNode *) lPtr->next;
	}
}


void
initializeMemQueue(memQueue * Q)
{
	_initializeMemQueue(Q, MINIMUM_ALLOCATED_NUM_BYTES);
}

float
numPagesInMemQueue(memQueue * Q)
{
	return ((float) Q->sizeInBytes / BLCKSZ);
}

int
ceilNumPagesInMemQueue(memQueue * Q)
{
	int			floor = ((int) (Q->sizeInBytes / BLCKSZ));

	if (floor > (int) floor)
		return ((int) floor + 1);
	else
		return (int) floor;
}

int
floorNumPagesInMemQueue(memQueue * Q)
{
	return ((int) (Q->sizeInBytes / BLCKSZ));
}

/*
 * Merges two tuple sets in the sense that the result is a tuple set that includes the tuples
 * from both tuple sets and the resulting tuple set is only join consistent and connected (JCC).
 * 
 * #define NO_MERGE				   0
 * #define LEFT_CONTAINED_RIGHT    1
 * #define RIGHT_CONTAINED_LEFT    2
 * #define MERGED_INTO_NEW_TSET    3
 * in all cases of contained or merged. update the status vars
 * visited, toVisit, relsOfTuples to the
  */
int
mergeTwoTSets(alg_fctx * fctx, TSet ltset, Datum *ltsetVals, char *ltsetNulls,
			  bool *ltsetDeformed, RBITS lrelsOfTuples, RBITS ltoVisit, RBITS lvisited, TSet rtset,
			  RBITS rrelsOfTuples, RBITS rtoVisit, RBITS rvisited,
	 TSet * mtset, RBITS * mrelsOfTuples, RBITS * mtoVisit, RBITS * mvisited)
{
	bytea	   *ltids = ltset->tids;
	bytea	   *rtids = rtset->tids;
	RBITS		intersection,
				tmp,
				tmp2;

	*ltsetDeformed = true;
	bool		attemptJCCUnify = false;
	TSet		unifyJCCResult;
	int			j = 0;
	RBITS		ltsetBlocked = ((~lrelsOfTuples) & lvisited);
	RBITS		rtsetBlocked = ((~rrelsOfTuples) & rvisited);

	/*
	 * if there are tuples in one tuple set that were blocked(visited but no
	 * tuple) then surely they CANNOT merge. Note, its first because it must
	 * be done in order. i used a logic that depends on the ORDER ORDER ORDER
	 * ORDER!
	 */
	if ((lrelsOfTuples & rtsetBlocked) || (rrelsOfTuples & ltsetBlocked));

	/*
	 * Lets check if there is an intersection of relations between the tuple
	 * sets. If there is an intersection, lets compare the rowids.
	 */
	else if ((intersection = (lrelsOfTuples & rrelsOfTuples)))
	{
		int			leftPos = 0;
		int			leftRelOid = 0;
		int			rightRelOid = 0;
		int			numRightTuples = numOfTuplesInTSet(fctx, rtids);

		leftRelOid = getTableIdOfRelsOidInTSet(fctx, ltids, leftPos);
		for (j = 0; j < numRightTuples; j++)
		{
			rightRelOid = getTableIdOfRelsOidInTSet(fctx, rtids, j);
			if (varNthBitFromLeft(intersection, rightRelOid))
			{
				while (leftRelOid != rightRelOid)
				{
					leftPos++;
					leftRelOid = getTableIdOfRelsOidInTSet(fctx, ltids, leftPos);
				}
				if (memcmp(VARDATA(ltids) + leftPos * (1 + sizeof(ItemPointerData))
						 ,VARDATA(rtids) + j * (1 + sizeof(ItemPointerData)),
						   1 + (sizeof(ItemPointerData))))
				{
					break;
				}
			}
			tmp = (~(0x0));
			if (!(intersection & (tmp >> (rightRelOid + 1))))
			{
				j = numRightTuples;
				break;
			}
		}
		if (j >= numRightTuples)
		{
			/*
			 * if the left side is contained in right side then don't do
			 * anything but consider it merged. So lets check if one contains
			 * the other:
			 */
			if ((rrelsOfTuples | lrelsOfTuples) == rrelsOfTuples)
			{
				*mrelsOfTuples = rrelsOfTuples;
				*mvisited = (~(rrelsOfTuples)) & (rvisited | lvisited);
				*mtoVisit = (~(rrelsOfTuples)) & (rtoVisit | ltoVisit)
					& (~(rvisited | lvisited));
				return RIGHT_CONTAINED_LEFT;
			}
			else if ((rrelsOfTuples | lrelsOfTuples) == lrelsOfTuples)
			{
				/* replace ltset with the rtset. */
				*mrelsOfTuples = lrelsOfTuples;
				*mvisited = (~(lrelsOfTuples)) & (rvisited | lvisited);
				*mtoVisit = (~(lrelsOfTuples)) & (rtoVisit | ltoVisit)
					& (~(rvisited | lvisited));
				return LEFT_CONTAINED_RIGHT;
			}
			else
			{
				/*
				 * Here we want to check if relations of tuples of one of the
				 * tuple sets that are not in the intesection, are connected
				 * to tuples of the other tuple set that are not in the
				 * intersection. Because if so, we cannot use dumb unify and
				 * we have to check also for JCC.
				 */
				j = 0;
				tmp2 = 0;
				/* tuples of rtset not in intersection */
				tmp = (~intersection) & rrelsOfTuples;

				/*
				 * tmp2 contains relations that relations of tuples not in
				 * intersection, are connected to.
				 */
				while (tmp)
				{
					if (varNthBitFromLeft(tmp, 0))
						tmp2 |= (fctx->bits_scheme_graph[j]);
					j++;
					tmp <<= 1;
				}

				/*
				 * we check if the tuples of rtset not in the intersections
				 * are connected to tuples of ltset not in the intersection. f
				 * not we continue, otherwise we will later attempt Unifying
				 * with JCC(ELSE CLAUSE).
				 */
				if (!((~intersection) & lrelsOfTuples & tmp2))
				{
					j = 0;
					tmp2 = 0;
					/* tuples of ltset not in intersection. */
					tmp = (~intersection) & lrelsOfTuples;
					while (tmp)
					{
						if (varNthBitFromLeft(tmp, 0))
							tmp2 |= (fctx->bits_scheme_graph[j]);
						j++;
						tmp <<= 1;
					}

					/*
					 * we check if the tuples of ltset not in the
					 * intersections are connected to tuples of rtset not in
					 * the intersection. if not we will unify the tuple sets
					 * without checks. otherwise we will proceed to unifying
					 * with checks (ELSE CLAUSE).
					 */
					if (!((~intersection) & rrelsOfTuples & tmp2))
					{
						unifyJCCResult = simpleUnify(fctx, ltset, ltsetVals, ltsetNulls, rtset);
						*mtset = unifyJCCResult;
						*mrelsOfTuples = lrelsOfTuples | rrelsOfTuples;
						*mvisited = (~(rrelsOfTuples)) & (rvisited | lvisited);
						*mtoVisit = (~(rrelsOfTuples)) & (rtoVisit | ltoVisit)
							& (~(rvisited | lvisited));
						return MERGED_INTO_NEW_TSET;
					}
					else
						attemptJCCUnify = true;
				}
				else
					attemptJCCUnify = true;
			}
		}
	}

	/*
	 * Otherwise, there was no intersection of relations:
	 * ------------------------------ Here we CANNOT merge if all the
	 * relations, either tuple set wants to visit, were blocked in the other
	 * tuple set. Note there can be lack of information so blocked only
	 * reflect the knowledge of visiting if we did try to visit.
	 */
	else if ((!(ltoVisit & (~rtsetBlocked))) && (!(rtoVisit & (~ltsetBlocked))));

	/*
	 * We will only try to unify if both tuple sets needs to visit a tuple in
	 * the other tuple set.
	 */
	else if ((ltoVisit & rrelsOfTuples) || (rtoVisit & lrelsOfTuples))
		attemptJCCUnify = true;
	if (attemptJCCUnify)
	{
		if ((unifyJCCResult =
			 unifyAndJCC(fctx, ltset, ltsetVals, ltsetNulls, rtset, ((~intersection)
				& ((lrelsOfTuples & rtoVisit) | (rrelsOfTuples & ltoVisit))),
						 lrelsOfTuples, rrelsOfTuples)) != NULL)
		{
			*mtset = unifyJCCResult;
			*mrelsOfTuples = lrelsOfTuples | rrelsOfTuples;
			*mvisited = (~(rrelsOfTuples)) & (rvisited | lvisited);
			*mtoVisit = (~(rrelsOfTuples)) & (rtoVisit | ltoVisit)
				& (~(rvisited | lvisited));
			return MERGED_INTO_NEW_TSET;
		}
	}
	return NO_MERGE;
}

/* attempt to merge the tuple sets provided in Q to the node provided. obviously
 * only a limited nuber of them can be actually merged since the resulting
 * tuple set must be join consistent and connected which also means a tuple set
 * can only holds at most one tuple from each relation. */
bool
mergeOutOfMemQueueIntoNode(alg_fctx * fctx, memQueue * Q, memQueueNode * node,
						   bool exitAtFirstMerge, bool assumeQueueMerged)
{
	RBITS		relsOfTuples = node->relsOfTuples;
	RBITS		toVisit = node->toVisit;
	RBITS		visited = node->visited;
	Datum	   *tsetVals = node->tset->v;
	char	   *tsetNulls = node->tset->n;
	bool		deformedTSet = false;
	int			result;
	RBITS		mrelsOfTuples,
				mtoVisit,
				mvisited;
	TSet		mtset = NULL;
	TSet		tset = node->tset;
	memQueueNode *ptr = Q->head;
	memQueueNode *prevPtr = ptr;
	bool		wasMerging = false;

	while (ptr != NULL)
	{
		if ((result = mergeTwoTSets(fctx, tset, tsetVals, tsetNulls,
					&deformedTSet, relsOfTuples, toVisit, visited, ptr->tset,
							   ptr->relsOfTuples, ptr->toVisit, ptr->visited,
							  &mtset, &mrelsOfTuples, &mtoVisit, &mvisited)))
		{
			switch (result)
			{
				case LEFT_CONTAINED_RIGHT:
					removeMemQueueNode(Q, prevPtr, ptr);
					wasMerging = true;
					relsOfTuples = mrelsOfTuples;
					toVisit = mtoVisit;
					visited = mvisited;
					break;
				case RIGHT_CONTAINED_LEFT:
					detachNodeFromMemQueueNode(prevPtr, ptr, (Q));
					freeTSet(fctx, tset);
					node->tset = ptr->tset;
					tset = ptr->tset;
					deformedTSet = false;
					wasMerging = true;
					relsOfTuples = ptr->relsOfTuples;
					toVisit = ptr->toVisit;
					visited = ptr->visited;
					pfree(ptr);
					ptr = prevPtr;
					break;
				case MERGED_INTO_NEW_TSET:
					removeMemQueueNode(Q, prevPtr, ptr);
					node->tset = mtset;
					tset = mtset;
					deformedTSet = false;
					wasMerging = true;
					relsOfTuples = mrelsOfTuples;
					toVisit = mtoVisit;
					visited = mvisited;
			}
		}
		if ((exitAtFirstMerge && wasMerging) || (assumeQueueMerged && (result == RIGHT_CONTAINED_LEFT)))
			break;
		getMemQueueNextElementAfterRemovalPtr(prevPtr, ptr, Q);
	}
	node->relsOfTuples = relsOfTuples;
	node->toVisit = toVisit;
	node->visited = visited;
	return wasMerging;
}

/* add a tuple set to a memqueue with added capabilities of filling the memqueue
 * nodes with all their needed information */
void
addTupleSetToMemQueue(alg_fctx * fctx, memQueue * Q, TSet tset)
{
	RBITS		relsOfTuples = 0;
	RBITS		toVisit = 0;
	bytea	   *b = tset->tids;
	int			j,
				relID;
	int			numTuples = numOfTuplesInTSet(fctx, b);

    /* here we generate information from the tuple set for the bit arrays
     * relsOfTuples - 1 for each relation that has a tuple in the tuple set.
     * toVisit - 1 for a relation we will potentially need to extend to (determined
     * using the scheme graph 
     * visited 1 for each relation we visited. since we have not visited yet any relation
     * for this new node then all the bits are 0. */
	for (j = 0; j < numTuples; j++)
	{
		relID = getTableIdOfRelsOidInTSet(fctx, b, j);
		varSetNthBitFromLeft(relsOfTuples, relID);
	}
	for (j = 0; j < fctx->numRels; j++)
		if ((varNthBitFromLeft(relsOfTuples, j)))
			toVisit |= ((~relsOfTuples) & (fctx->bits_scheme_graph[j]));
	memQueueNode *newNode = palloc(sizeof(memQueueNode));

	newNode->tset = tset;
	newNode->relsOfTuples = relsOfTuples;
	newNode->toVisit = toVisit;
	newNode->visited = 0;
	newNode->next = NULL;
	if (Q->head == NULL)
		Q->head = newNode;
	if (Q->tail != NULL)
		Q->tail->next = (struct memQueueNode *) newNode;
	Q->tail = newNode;
	Q->numElements++;
	Q->sizeInBytes += tset->size;
}

/* remove from the left memqueue, nodes
 * , that are subsumed in the right provided tset (tuple set) */
bool
removeLeftMemQueueNodesSubsumedInRightTSet(alg_fctx * fctx, memQueue * Q, TSet tset)
{
	memQueueNode *ptr = Q->head;
	memQueueNode *prev = ptr;
	bool		wasRemoval = false;

	while (ptr != NULL)
	{
		if (isLeftSubsumedRightTset(fctx, ptr->tset, tset))
		{
			wasRemoval = true;
			removeMemQueueNode(Q, prev, ptr);
		}
		getMemQueueNextElementAfterRemovalPtr(prev, ptr, Q);
	}
	return wasRemoval;
}

/* remove nodes from the left memqueue that are subsumbed in the right
 * memqueue.*/
void
removeLeftMemQueueNodesSubsumedInRightMemQueueNodes(alg_fctx * fctx, memQueue * QLeft,
													memQueue * QRight)
{
    
    /* Not protected against self, use mergeMemQueueWithSelf instead */
	Assert(QLeft != QRight);
    
	memQueueNode *ptr = QRight->head;

	while (ptr != NULL)
	{
		removeLeftMemQueueNodesSubsumedInRightTSet(fctx, QLeft, ptr->tset);
		ptr = (memQueueNode *) ptr->next;
	}
}

/* clear a memory queue. release its resources */
void
clearMemQueue(alg_fctx * fctx, memQueue * Q, bool free_tsets)
{
	memQueueNode *ptr = Q->head;
	memQueueNode *prevPtr = ptr;

	while (ptr != NULL)
	{
		if (free_tsets)
			freeTSet(fctx, ptr->tset);
		prevPtr = ptr;
		ptr = (memQueueNode *) ptr->next;
		pfree(prevPtr);
	}
	Q->tail = NULL;
	Q->head = NULL;
	Q->numElements = 0;
	Q->sizeInBytes = 0;
}

/* initialize a tuplestore (represented as a storage queue) */
void
SQInitialize(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	Q->tuplestorage = tuplestore_begin_heap(false, false, (int) (BLCKSZ / 1024));
#ifdef PG_MODULE_MAGIC
	Q->slot = MakeSingleTupleTableSlot(fctx->tupleSetDesc);
#endif   /* */
	MemoryContextSwitchTo(oldcontext);
	Q->numElements = 0;
	Q->sizeInBytes = 0;
	Q->lastReadPosition = 0;
	Q->markedLastReadPosition = -1;
	Q->markedNumElements = -1;
	Q->markedSizeInBytes = -1;
	Q->markedMemOverflowTuple = NULL;
	Q->memOverflowTuple = NULL;
}

/* release the resoures of the tuple store (represented as a storage queue) */
void
SQFinish(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	if (Q->tuplestorage != NULL)
		tuplestore_end(Q->tuplestorage);
	MemoryContextSwitchTo(oldcontext);
	Q->tuplestorage = NULL;
	Q->numElements = 0;
	Q->sizeInBytes = 0;
	Q->markedLastReadPosition = -1;
	Q->markedNumElements = -1;
	Q->markedSizeInBytes = -1;
	Q->markedMemOverflowTuple = NULL;
	Q->lastReadPosition = 0;
	if (Q->memOverflowTuple != NULL)
		heap_freetuple(Q->memOverflowTuple);
	if (Q->markedMemOverflowTuple != NULL)
		heap_freetuple(Q->markedMemOverflowTuple);
}

/*
 * the caller must return the tuple he took out.
  */
void
SQSetMemOverflowTuple(storageQueue * Q, HeapTuple tuple)
{
	Q->memOverflowTuple = tuple;
	Q->lastReadPosition--;
	Q->sizeInBytes += tupleSize(tuple);
}

/* add a heaptuple to the storage queue (tuplestore) */
void
SQAddElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q, HeapTuple tuple)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	tuplestore_puttuple(Q->tuplestorage, tuple);
	MemoryContextSwitchTo(oldcontext);
	Q->numElements++;
	Q->sizeInBytes += tupleSize(tuple);
}

/* get the next element from the storage queue (tuple store)*/
HeapTuple
SQGetNextElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q, bool *should_free,
				 bool copytuple)
{
	Q->lastReadPosition++;
	HeapTuple	tmpTuple = NULL;

	if (Q->memOverflowTuple != NULL)
	{
		*should_free = true;
		tmpTuple = Q->memOverflowTuple;
		Q->memOverflowTuple = NULL;
		Q->sizeInBytes -= tupleSize(tmpTuple);
		return tmpTuple;
	}
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);
	HeapTuple	tuple = NULL;
/*compatibility between 8.1 and 8.2 postgresql versions */
#ifdef PG_MODULE_MAGIC
	ExecClearTuple(Q->slot);
	if (tuplestore_gettupleslot(Q->tuplestorage, true, Q->slot))
	{
		tuple = ExecFetchSlotTuple(Q->slot);
	}
	else
       elog(ERROR, "At %s(%d) - no tuple returned. We do not check here for end of queue", __FILE__, __LINE__);
#else							/* */
	tuple = tuplestore_gettuple(Q->tuplestorage, true, should_free);
#endif   /* */
	MemoryContextSwitchTo(oldcontext);
	if (copytuple)
	{
		tmpTuple = (HeapTuple) heap_copytuple(tuple);
#ifndef PG_MODULE_MAGIC
		if (*should_free)
			heap_freetuple(tuple);
#endif   /* */
		*should_free = true;
		Q->sizeInBytes -= tupleSize(tmpTuple);
		return tmpTuple;
	}
	else
	{
		Q->sizeInBytes -= tupleSize(tuple);
		return tuple;
	}
}

/* go back to the first element in the queue */
void
SQRewindToFirstElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	tuplestore_rescan(Q->tuplestorage);
	MemoryContextSwitchTo(oldcontext);
	Q->lastReadPosition = 0;
}

/* mark the current position in the tuplestore */
void
SQMarkPosition(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	tuplestore_markpos(Q->tuplestorage);
	MemoryContextSwitchTo(oldcontext);
	Q->markedLastReadPosition = Q->lastReadPosition;
	Q->markedNumElements = Q->numElements;
	Q->markedSizeInBytes = Q->sizeInBytes;
	if (Q->memOverflowTuple != NULL)
		Q->markedMemOverflowTuple = heap_copytuple(Q->memOverflowTuple);
	else
		Q->markedMemOverflowTuple = NULL;
}

/* restore a previously saved position in the tuple store */
void
SQRestorePosition(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q)
{
	MemoryContext oldcontext = MemoryContextSwitchTo(fctx->per_query_ctx);

	tuplestore_restorepos(Q->tuplestorage);
	MemoryContextSwitchTo(oldcontext);
	Q->lastReadPosition = Q->markedLastReadPosition;
	Q->numElements = Q->markedNumElements;
	Q->sizeInBytes = Q->markedSizeInBytes;
	if (Q->memOverflowTuple)
	{
		heap_freetuple(Q->memOverflowTuple);
		Q->memOverflowTuple = NULL;
	}
	if (Q->markedMemOverflowTuple != NULL)
		Q->memOverflowTuple = Q->markedMemOverflowTuple;
	Q->markedLastReadPosition = -1;
	Q->markedNumElements = -1;
	Q->markedSizeInBytes = -1;
	Q->markedMemOverflowTuple = NULL;
}
