/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#ifndef QUEUESFUNCS_H
#define QUEUESFUNCS_H

#define MINIMUM_ALLOCATED_NUM_PAGES 2
#define MINIMUM_ALLOCATED_NUM_BYTES (MINIMUM_ALLOCATED_NUM_PAGES*BLCKSZ)

/* 
 * not all 64 bits systems or even 32 with limited 64 bits support can support
 * 64 bits bitwise operations!!!!! 
 * */
#ifdef BITS_64_BITWISE_SUPPORT
  #define firstBitFromLeftInHex 0x8000000000000000
#else
  #define firstBitFromLeftInHex 0x80000000
#endif

/* helper constants for merging two tuple sets */
#define NO_MERGE				0
#define LEFT_CONTAINED_RIGHT	1
#define RIGHT_CONTAINED_LEFT	2
#define MERGED_INTO_NEW_TSET	3

#include "utils/palloc.h"
#include "queues.h"
#include "algstructs.h"

/* n is starting from zero! */
#define nthBitFromLeft(n)\
(firstBitFromLeftInHex >> n)

#define varNthBitFromLeft(var,n)\
((var) & (nthBitFromLeft(n)))

#define varSetNthBitFromLeft(var,n)\
(var = (var) | (nthBitFromLeft(n)))

#define varUnSetNthBitFromLeft(var,n)\
(var = (var) & (~(nthBitFromLeft(n))))

/*replace the tuple set of a queue node */
#define replaceNodeTSet(fctx,node,retset)\
do\
{\
	freeTSet(fctx, (node)->tset);\
		(node)->tset = retset;\
} while (0)

/* free the resources of a memory queue node */
#define freeMemQueueNode(node)\
do\
{\
	freeTSet(fctx, node->tset);	\
		pfree(node);\
		node = NULL;\
} while (0)

/* add a node to a memory queue*/
#define addNodeToMemQueue(node,Q) \
do\
{\
	if ((Q)->head == NULL)\
		(Q)->head = (node); \
		else\
		(Q)->tail->next = (struct memQueueNode *) (node);\
		(node)->next = NULL;\
		(Q)->numElements++;\
		(Q)->sizeInBytes += (node)->tset->size;\
		(Q)->tail = (node);\
	} while (0)

/*node is modified to detached node.
 * the head node is detached from the queue.*/
#define detachHeadNodeFromMemQueue(node,Q)\
do\
	{\
		if ((Q)->head != NULL)\
		{\
			node = (Q)->head;\
				(Q)->head = (memQueueNode *) node->next;\
				(Q)->numElements--;	\
				(Q)->sizeInBytes -= (node)->tset->size;	\
				if ((Q)->head == NULL)\
				(Q)->tail = NULL;\
		} \
			else\
			node = NULL;\
	} while (0)

/* detach an arbitrary node from a memory queue. you must provide
 * the previous memory queue node or the same if it is the head
 * node */
#define detachNodeFromMemQueueNode(prevNode,node,Q)\
do\
{\
	if ((node) == (prevNode))\
	{\
			if ((prevNode) == (Q)->head)\
				(Q)->head = (memQueueNode *) (prevNode)->next;\
			if ((prevNode) == (Q)->tail)\
				(Q)->tail = NULL;\
			(prevNode) = NULL;\
	}\
	else\
	{\
			(prevNode)->next = (node)->next;\
			if ((node) == (Q)->tail)\
				(Q)->tail = (prevNode);\
	} \
		(Q)->numElements--;\
		(Q)->sizeInBytes -= (node)->tset->size;	\
} while (0)

/* removes a mem queue node, as opposed to detaching we also
 * free the removed node resources */
#define removeMemQueueNode(Q, prevPtr, ptr)\
do\
{\
	detachNodeFromMemQueueNode((prevPtr), (ptr), (Q));\
		freeMemQueueNode(ptr);\
		(ptr) = (prevPtr);\
} while (0)

/* after using the above remove memory queue, get the next memory queue node.
 * you must use this macro otherwise you will jump over nodes */
#define getMemQueueNextElementAfterRemovalPtr(prevNodePtr,nodePtr,Q)\
do\
{\
	if (!prevNodePtr)\
	{\
			nodePtr = getMemQueueFirstElementPtr((Q));\
			prevNodePtr = nodePtr;\
	}\
	else if (nodePtr)\
	{\
			prevNodePtr = nodePtr;\
			nodePtr = getMemQueueNextElementPtr(nodePtr);\
	}\
} while (0)


/* get the tuple set of a memory queue node */
#define getMemQueueElement(ptr)\
(ptr->tset)


#define getMemQueueNextElementPtr(ptr)\
((memQueueNode *) (ptr->next))

#define getMemQueueFirstElementPtr(Q)\
((Q)->head)

#define getMemQueueLastElementPtr(Q)\
((Q)->tail)

#define getMemQueueNumElements(Q)\
((Q)->numElements)

#define getMemQueueSize(Q)\
((Q)->sizeInBytes)

#define isMemQueueEmpty(Q)\
(((Q)->numElements) <= 0)

#define SQGetNumElements(Q)\
((Q)->numElements)

#define numPagesInSQ(Q)\
((Q)->sizeInBytes / BLCKSZ)

/* check if we reached the end of the tuple store by using the counter instead of directly using
 * the tuple store facilities */
#define SQEof(Q)\
((Q)->numElements <= (Q)->lastReadPosition)

#define SQNotEmpty(Q)\
((Q)->numElements > (Q)->lastReadPosition)

extern void _initializeMemQueue(memQueue * Q, int allocatedBytes);
extern void removeDuplicatesMemQueueWithSelf(alg_fctx * fctx, memQueue * Q);
extern void removeSubsumptionsMemQueueWithSelf(alg_fctx * fctx, memQueue * Q);
extern void initializeMemQueue(memQueue * Q);
extern float numPagesInMemQueue(memQueue * Q);
extern int	floorNumPagesInMemQueue(memQueue * Q);
extern int	ceilNumPagesInMemQueue(memQueue * Q);
extern bool mergeOutOfMemQueueIntoNode(alg_fctx * fctx, memQueue * Q, memQueueNode * node,
						   bool exitAtFirstMerge, bool assumeQueueMerged);
extern void addTupleSetToMemQueue(alg_fctx * fctx, memQueue * Q, TSet tset);
extern bool removeLeftMemQueueNodesSubsumedInRightTSet(alg_fctx * fctx
										   ,memQueue * Q, TSet tset);
extern void removeLeftMemQueueNodesSubsumedInRightMemQueueNodes(alg_fctx * fctx
									   ,memQueue * QLeft, memQueue * QRight);
extern void clearMemQueue(alg_fctx * fctx, memQueue * Q, bool free_tsets);
extern void SQInitialize(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q);
extern void SQFinish(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q);
extern void SQSetMemOverflowTuple(storageQueue * Q, HeapTuple tuple);
extern void SQAddElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q, HeapTuple tuple);
extern HeapTuple SQGetNextElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q, bool *should_free,
				 bool copytuple);
extern void SQRewindToFirstElement(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q);
extern void SQMarkPosition(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q);
extern void SQRestorePosition(alg_fctx * fctx, FuncCallContext *funcctx, storageQueue * Q);

#endif   /* QUEUESFUNCS_H */
