/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#include "algstructs.h"
#include "tset.h"
#include "tsetfuncs.h"
#include "algutils.h"
#include "queues.h"
#include "queuesfuncs.h"

/* for backward compatibilty */
#ifndef HeapTupleHeaderGetNatts
# define HeapTupleHeaderGetNatts(tup) ((tup)->t_natts)
#endif

/* free a tuple set and its datums */
void
freeTSet(alg_fctx * fctx, TSet tset)
{
	if (tset->t == NULL)
	{
		int			i;

		for (i = 0; i < NUMATT; i++)
			if (tset->v[i] != (Datum) NULL)
				myDatumFree(tset->v[i], fctx->tupleSetDesc->attrs[i]->attbyval,
							fctx->tupleSetDesc->attrs[i]->attlen);
	}
	else
		heap_freetuple(tset->t);
	if (tset->v != NULL)
	{
		pfree(tset->v);
		pfree(tset->n);
	}
	pfree(tset);
}

/* create an empty TSet structure */
TSet
newEmptyTSet(alg_fctx * fctx)
{
	TSet		tset = ((TSet) palloc(sizeof(deformedTSet)));
	bytea	   *b = (bytea *) palloc(VARHDRSZ);

	SET_VARSIZE(b, VARHDRSZ);
	tset->v = NULL;
	tset->n = NULL;
	Datum		tsetVals[NUMATT];
	char		tsetNulls[NUMATT];
	int			i;

	for (i = FIRSTATT; i < NUMATT; i++)
		tsetNulls[i] = 'n';
	tsetNulls[LABELS] = ' ';
	tsetVals[LABELS] = (Datum) b;
	tset->t = heap_formtuple(fctx->tupleSetDesc, tsetVals, tsetNulls);
	tset->size = tupleSize(tset->t);
	bool		isnull = false;

	tset->tids = (bytea *) fastgetattr(tset->t, LABELS_ALIGNED, fctx->tupleSetDesc, &isnull);
	return tset;
}

/* set a datum value to a tuple set at attribute at position.
 * we also compy the datum value. */
Datum
setTSetValue(alg_fctx * fctx, TSet tset, Datum value, int position)
{
	if ((tset->v[position] != (Datum) NULL) && (tset->t == NULL))
	{
		myDatumFree(tset->v[position], fctx->tupleSetDesc->attrs[position]->attbyval,
					fctx->tupleSetDesc->attrs[position]->attlen);
		tset->size -= datumGetSize(tset->v[position], fctx->tupleSetDesc->attrs[position]->attbyval,
								fctx->tupleSetDesc->attrs[position]->attlen);
	}
	tset->v[position] = myDatumCopy(value, fctx->tupleSetDesc->attrs[position]->attbyval,
								fctx->tupleSetDesc->attrs[position]->attlen);
	tset->size += datumGetSize(value, fctx->tupleSetDesc->attrs[position]->attbyval,
							   fctx->tupleSetDesc->attrs[position]->attlen);
	return tset->v[position];
}

/* same as setTSetValue but we don't copy the datum so do not free it */
Datum
setTSetValueNoCopy(alg_fctx * fctx, TSet tset, Datum value, int position)
{
	if ((tset->v[position] != (Datum) NULL) && (tset->t == NULL))
	{
		myDatumFree(tset->v[position], fctx->tupleSetDesc->attrs[position]->attbyval,
					fctx->tupleSetDesc->attrs[position]->attlen);
		tset->size -= datumGetSize(tset->v[position], fctx->tupleSetDesc->attrs[position]->attbyval,
								fctx->tupleSetDesc->attrs[position]->attlen);
	}
	tset->v[position] = value;
	tset->size += datumGetSize(value, fctx->tupleSetDesc->attrs[position]->attbyval,
							   fctx->tupleSetDesc->attrs[position]->attlen);
	return tset->v[position];
}

/* here we create a new empty tset but without even its heaptuple */
TSet
newTSet_(alg_fctx * fctx)
{
	TSet		newtset = (TSet) palloc(sizeof(deformedTSet));

	newtset->t = NULL;
	newtset->size = 0;
	newtset->v = (Datum *) palloc(NUMATT * sizeof(Datum));
	newtset->n = (char *) palloc(NUMATT * sizeof(char));
	int			i = 0;

	for (i = 0; i < NUMATT; i++)
	{
		newtset->v[i] = (Datum) NULL;
		newtset->n[i] = 'n';
	}
	newtset->tids = NULL;
	return newtset;
}

/*new empty DTuple  numAtt is the number of attributes in the tuple*/
DTuple
newDTuple(alg_fctx * fctx, int numAtt)
{
	DTuple		tuple = (DTuple) palloc(sizeof(deformedTuple));

	tuple->t = NULL;
	tuple->deformed = false;
	tuple->v = (Datum *) palloc((numAtt + DTupleTupleHeaderSize) * sizeof(Datum));
	tuple->n = (char *) palloc((numAtt + DTupleTupleHeaderSize) * sizeof(char));
	return tuple;
}

/*is the attribute in pos in the tuple set Null? */
bool
tsetAttIsNull(TSet tset, int pos)
{
	if (tset->v != NULL)
		return (tset->n[pos] == 'n');
	else
		return heap_attisnull(tset->t, pos + 1);
}

/* new TSet structure from a heaptuple tuple set */
TSet
newTSet(alg_fctx * fctx, HeapTuple tupleTSet)
{
	TSet		newtset = (TSet) palloc(sizeof(deformedTSet));

	newtset->size = tupleSize(tupleTSet);
	newtset->t = tupleTSet;
	newtset->v = (Datum *) palloc(NUMATT * sizeof(Datum));
	newtset->n = (char *) palloc(NUMATT * sizeof(char));
	heap_deformtuple(tupleTSet, fctx->tupleSetDesc, newtset->v, newtset->n);
	newtset->tids = (bytea *) newtset->v[LABELS];
	return newtset;
}

/* new TSet structure from a heaptuple but we do not deform the tuple */
TSet
newTSet___(alg_fctx * fctx, HeapTuple tupleTSet)
{
	TSet		newtset = (TSet) palloc(sizeof(deformedTSet));

	newtset->size = tupleSize(tupleTSet);
	newtset->t = tupleTSet;
	newtset->v = NULL;
	newtset->n = NULL;
	bool		isnull;

	newtset->tids = (bytea *) fastgetattr(tupleTSet, LABELS_ALIGNED, fctx->tupleSetDesc, &isnull);
	return newtset;
}

/* new completly empty TSet structure */
TSet
newTSet____(alg_fctx * fctx)
{
	TSet		newtset = (TSet) palloc(sizeof(deformedTSet));

	newtset->size = 0;
	newtset->t = NULL;
	newtset->v = NULL;
	newtset->n = NULL;
	newtset->tids = NULL;
	return newtset;
}

/* print a tuple set labels into a string */
char *
printTSetToCStr(alg_fctx * fctx, bytea *b)
{
	int			numSets = numOfTuplesInTSet(fctx, b);
	char	   *tsetStr;

	if (numSets > 0)
	{
		int			i;

		tsetStr = palloc(1600 * sizeof(char));
		char	   *to = tsetStr;
		ItemPointerData *ctid = palloc(sizeof(ItemPointerData));

		for (i = 0; i < numSets; i++)
		{
			memcpy(ctid, VARDATA(b) + i * (sizeof(ItemPointerData) + 1) + 1, sizeof(ItemPointerData));
			char		tempStr[100];

			sprintf(tempStr, "[TRID:%d, ROWID:%d,%d,%d]"
				,getTableIdOfRelsOidInTSet(fctx, b, i), ctid->ip_blkid.bi_hi,
					ctid->ip_blkid.bi_lo, ctid->ip_posid);
			to = (char *) stpcpy(to, tempStr);
		}
		pfree(ctid);
	}
	else
	{
		tsetStr = palloc(8 * sizeof(char));
		char	   *to = tsetStr;

		to = (char *) stpcpy(to, "[]");
	}
	return tsetStr;
}

/* Print a tuple set contents to the console (used for debugging)*/
void
printTSet(alg_fctx * fctx, HeapTuple tset)
{
	Assert(tset != NULL);

	Datum		values[NUMATT];
	char		nulls[NUMATT];

	heap_deformtuple(tset, fctx->tupleSetDesc, values, nulls);
	bytea	   *b = (bytea *) values[LABELS];
	char	   *tsetStr = printTSetToCStr(fctx, b);
	bool		tofree = true;
	int			bufSize = tupleSize(tset);
	char	   *buf = (char *) palloc(bufSize);
	char	   *to = buf;

	buf = addAndExpandStringBy(buf, &bufSize, tsetStr, BLCKSZ, &to, &tofree);
	pfree(tsetStr);
	int			i = 0;

	for (i = FIRSTATT; i < fctx->tupleSetNatt; i++)
		if (nulls[i] == ' ')
		{
			buf = addAndExpandStringBy(buf, &bufSize, ",", BLCKSZ, &to, &tofree);
			buf = addAndExpandStringBy(buf, &bufSize, SPI_getvalue(tset, fctx->tupleSetDesc, i + 1)
									   ,BLCKSZ, &to, &tofree);
		}
		else
			buf = addAndExpandStringBy(buf, &bufSize, ", NULL", BLCKSZ, &to, &tofree);
	/* PRINT TO SCREEN HERE (DON'T DELETE!!). */
	elog(INFO, "%s", buf);
	if (tofree)
		pfree(buf);
}

/* print a deformed tuple set to the console (used for debugging purposes */
void
printDTSet(alg_fctx * fctx, TSet tset)
{
	Assert(tset != NULL);

	Datum	   *values;
	char	   *nulls;

	values = tset->v;
	nulls = tset->n;
	bytea	   *b = (bytea *) values[LABELS];
	char	   *tsetStr = printTSetToCStr(fctx, b);
	bool		tofree = true;
	int			bufSize = 10 * (tset->size + 1);
	char	   *buf = (char *) palloc(bufSize);
	char	   *to = buf;

	buf = addAndExpandStringBy(buf, &bufSize, tsetStr, BLCKSZ, &to, &tofree);
	pfree(tsetStr);
	int			i = 0;
	char		value[1000];

	for (i = FIRSTATT; i < fctx->tupleSetNatt; i++)
		if (nulls[i] == ' ')
		{
			buf = addAndExpandStringBy(buf, &bufSize, ",", BLCKSZ, &to, &tofree);
			sprintf(value, "%d", (int) values[i]);
			buf = addAndExpandStringBy(buf, &bufSize, value
									   ,BLCKSZ, &to, &tofree);
		}
		else
			buf = addAndExpandStringBy(buf, &bufSize, ", NULL", BLCKSZ, &to, &tofree);
	/* PRINT TO SCREEN HERE (DON'T DELETE!!). */
	elog(INFO, "%s", buf);
	if (tofree)
		pfree(buf);
}

/*print a deformed tuple to the console (used for debugging purposes) */
void
printDTuple(alg_fctx * fctx, DTuple dtuple, int tupleRelID)
{
	Assert (dtuple != NULL);

	Datum	   *values = dtuple->v;
	char	   *nulls = dtuple->n;
	bool		tofree = true;
	int			bufSize = 256 * sizeof(char);
	char	   *buf = (char *) palloc(bufSize);
	char	   *to = buf;
	int			i = 0;
	char		value[1000];

	sprintf(value, "RelID[%d]", tupleRelID);
	buf = addAndExpandStringBy(buf, &bufSize, value, BLCKSZ, &to, &tofree);
	for (i = 0; i < fctx->relsNAtt[tupleRelID]; i++)
		if (nulls[i] == ' ')
		{
			buf = addAndExpandStringBy(buf, &bufSize, ",", BLCKSZ, &to, &tofree);
			sprintf(value, "%d", (int) values[i]);
			buf = addAndExpandStringBy(buf, &bufSize, value
									   ,BLCKSZ, &to, &tofree);
		}
		else
			buf = addAndExpandStringBy(buf, &bufSize, ", NULL", BLCKSZ, &to, &tofree);
	/* PRINT TO SCREEN HERE (DON'T DELETE!!). */
	elog(INFO, "%s", buf);
	if (tofree)
		pfree(buf);
}

/*
 * Creates a bytea containing a the set of Rowids of the tuple set.
 * including the tableOidinRelsOID preceding each one.
 * we use one byte for each tableOidinRelsOID i.e there could be only
 * 256 tables (more than enough).
 * remember to FREE the returned value!
  */
bytea *
createNewTSetRowIDs(alg_fctx * fctx, ItemPointer tid, int tupleInRelsOid)
{
	bytea	   *b = palloc(VARHDRSZ + 1 + sizeof(ItemPointerData));

	SET_VARSIZE(b, 1 + sizeof(ItemPointerData) + VARHDRSZ);
	VARDATA(b)[0] = (char) tupleInRelsOid;
	memcpy(VARDATA(b) + 1, tid, sizeof(ItemPointerData));
	return b;
}

/* add a rowid (TableID,CTID) to a bytea array which is the LABELS attribute
 * in a tuple set ). The baseSet is the set of rowids we wish to add the rowid.
 * The setContainingTRID is the set where the rowid we wish to add to the baseset
 * resides. tupleInRelsOid is the position in the setContainingTRID where the rowid
 * resides. You must specify tupleInRelsOid precicely otherwise it is a fatal error.*/
bytea *
addRowIDtoTSet__(alg_fctx * fctx, bytea *baseSet, bytea *setContainingTRID
				 ,int tupleInRelsOid)
{
	int			i = 0;
	int			SCTRIDnumTSets = numOfTuplesInTSet(fctx, setContainingTRID);
	int			SCTRIDpos = -1;

	for (i = 0; i < SCTRIDnumTSets; i++)
		if (tupleInRelsOid == getTableIdOfRelsOidInTSet(fctx, setContainingTRID, i))
		{
			SCTRIDpos = i;
			break;
		}
	
    Assert(SCTRIDpos >= 0);
    
	int			biggerPosition = -1;
	int			numTSets = numOfTuplesInTSet(fctx, baseSet);
	int			relID = -1;

	for (i = 0; i < numTSets; i++)
	{
		relID = getTableIdOfRelsOidInTSet(fctx, baseSet, i);
		if (relID == tupleInRelsOid)
		{
			bytea	   *sameb = palloc(VARSIZE(baseSet));

			memcpy(sameb, baseSet, VARSIZE(baseSet));
			return sameb;
		}
		else if (relID > tupleInRelsOid)
		{
			biggerPosition = i;
			break;
		}
	}
	if (biggerPosition < 0)
		biggerPosition = numTSets + 1;
	else
		biggerPosition++;
	bytea	   *outb = palloc(VARSIZE(baseSet) + 1 + sizeof(ItemPointerData));

	memcpy(outb, baseSet, VARHDRSZ + (biggerPosition - 1) * (1 + sizeof(ItemPointerData)));
	SET_VARSIZE(outb, VARSIZE(baseSet) + 1 + sizeof(ItemPointerData));
	VARDATA(outb)[(biggerPosition - 1) * (1 + sizeof(ItemPointerData))] = (char) tupleInRelsOid;
	memcpy(VARDATA(outb) + (biggerPosition - 1) * (1 + sizeof(ItemPointerData)) + 1
		   ,VARDATA(setContainingTRID) + SCTRIDpos * (1 + sizeof(ItemPointerData)) + 1
		   ,sizeof(ItemPointerData));
	memcpy(VARDATA(outb) + (biggerPosition) * (1 + sizeof(ItemPointerData))
	 ,VARDATA(baseSet) + (biggerPosition - 1) * (1 + sizeof(ItemPointerData))
		   ,VARSIZE(baseSet) - VARHDRSZ - (biggerPosition - 1) * (1 + sizeof(ItemPointerData)));
	return outb;
}


/* add a rowid (tableid,ctid) where tid is the CTID and tupleInRelsOid is the tableid 
 * to the LABELS attribute of a tuple set which is b */
bytea *
addRowIDtoTSet_(alg_fctx * fctx, bytea *b, ItemPointer tid, int tupleInRelsOid)
{
	int			i = 0;
	int			biggerPosition = -1;
	int			numTSets = numOfTuplesInTSet(fctx, b);
	int			relID = -1;

	for (i = 0; i < numTSets; i++)
	{
		relID = getTableIdOfRelsOidInTSet(fctx, b, i);
		if (relID == tupleInRelsOid)
		{
			bytea	   *sameb = palloc(VARSIZE(b));

			memcpy(sameb, b, VARSIZE(b));
			return sameb;
		}
		else if (relID > tupleInRelsOid)
		{
			biggerPosition = i;
			break;
		}
	}
	if (biggerPosition < 0)
		biggerPosition = numTSets + 1;
	else
		biggerPosition++;
	bytea	   *outb = palloc(VARSIZE(b) + 1 + sizeof(ItemPointerData));

	memcpy(outb, b, VARHDRSZ + (biggerPosition - 1) * (1 + sizeof(ItemPointerData)));
	SET_VARSIZE(outb, VARSIZE(b) + 1 + sizeof(ItemPointerData));
	VARDATA(outb)[(biggerPosition - 1) * (1 + sizeof(ItemPointerData))] = (char) tupleInRelsOid;
	memcpy(VARDATA(outb) + (biggerPosition - 1) * (1 + sizeof(ItemPointerData)) + 1
		   ,tid, sizeof(ItemPointerData));
	memcpy(VARDATA(outb) + (biggerPosition) * (1 + sizeof(ItemPointerData))
		   ,VARDATA(b) + (biggerPosition - 1) * (1 + sizeof(ItemPointerData))
		   ,VARSIZE(b) - VARHDRSZ - (biggerPosition - 1) * (1 + sizeof(ItemPointerData)));
	return outb;
}


bytea *
addRowIDtoTSet(alg_fctx * fctx, TSet tset, ItemPointer tid, int tupleInRelsOid)
{
	return ((bytea *) addRowIDtoTSet_(fctx, tset->tids
									  ,tid, tupleInRelsOid));
}


/* merge two LABELS attribute of tuple sets. since these are sets we must remove
 * duplicates. */
bytea *
mergeRowIDs(alg_fctx * fctx, bytea *bLeft, bytea *bRight)
{
	bytea	   *tempOutb = palloc(VARSIZE(bLeft) + VARSIZE(bRight) - VARHDRSZ);
	int			maxl = numOfTuplesInTSet(fctx, bLeft);
	int			maxr = numOfTuplesInTSet(fctx, bRight);
	int			l = 0,
				r = 0;
	int			lTID = -1,
				rTID = -1;
	int			pos = 0;

	SET_VARSIZE(tempOutb, VARSIZE(bLeft) + VARSIZE(bRight) - VARHDRSZ);
	while (true)
	{
		lTID = getTableIdOfRelsOidInTSet(fctx, bLeft, l);
		rTID = getTableIdOfRelsOidInTSet(fctx, bRight, r);
		if (lTID < rTID)
		{
			memcpy(VARDATA(tempOutb) + pos * (1 + sizeof(ItemPointerData))
				   ,VARDATA(bLeft) + l * (1 + sizeof(ItemPointerData)), (1 + sizeof(ItemPointerData)));
			l++;
		}
		else
		{
			memcpy(VARDATA(tempOutb) + pos * (1 + sizeof(ItemPointerData))
				   ,VARDATA(bRight) + r * (1 + sizeof(ItemPointerData)), (1 + sizeof(ItemPointerData)));
			if (lTID == rTID)
			{
				r++;
				l++;
			}
			else
				r++;
		}
		pos++;
		if ((l == maxl) && (r == maxr))
		{
			break;
		}
		else if (l == maxl)
		{
			memcpy(VARDATA(tempOutb) + pos * (1 + sizeof(ItemPointerData))
				   ,VARDATA(bRight) + r * (1 + sizeof(ItemPointerData))
				   ,VARSIZE(bRight) - VARHDRSZ - r * (1 + sizeof(ItemPointerData)));
			pos += maxr - r;
			break;
		}
		else if (r == maxr)
		{
			memcpy(VARDATA(tempOutb) + pos * (1 + sizeof(ItemPointerData))
				   ,VARDATA(bLeft) + l * (1 + sizeof(ItemPointerData))
				   ,VARSIZE(bLeft) - VARHDRSZ - l * (1 + sizeof(ItemPointerData)));
			pos += maxl - l;
			break;
		}
	}

	SET_VARSIZE(tempOutb, VARHDRSZ + pos * (1 + sizeof(ItemPointerData)));
	repalloc(tempOutb, VARSIZE(tempOutb));
	return tempOutb;
}

/* are two tuple sets are equal. simply check if their lables attribute is equal. */
bool
areTSetsEqual(alg_fctx * fctx, TSet tset1, TSet tset2)
{
	if (VARSIZE(tset1->tids) != VARSIZE(tset2->tids))
		return false;
	return (memcmp(tset1->tids, tset2->tids, VARSIZE(tset1->tids)) == 0);
}

/* check if the left tuple set is subsumed (their lables attribute which is a set
 * is contained in the other) in the right tuple set */
bool
isLeftSubsumedRightTset(alg_fctx * fctx, TSet left, TSet right)
{
	bytea	   *bLeft = left->tids;
	bytea	   *bRight = right->tids;

	if (VARSIZE(bLeft) > VARSIZE(bRight))
		return false;
	int			i = 0;
	int			leftPos = 0;
	int			leftRelOid = 0;
	int			rightRelOid = 0;
	int			numLeftTuples = numOfTuplesInTSet(fctx, bLeft);
	int			numRightTuples = numOfTuplesInTSet(fctx, bRight);

	leftRelOid = getTableIdOfRelsOidInTSet(fctx, bLeft, leftPos);
	for (i = 0; i < numRightTuples; i++)
	{
		rightRelOid = getTableIdOfRelsOidInTSet(fctx, bRight, i);
		if (leftRelOid == rightRelOid && (memcmp(VARDATA(bLeft) + leftPos * (1 + sizeof(ItemPointerData))
												 ,VARDATA(bRight) + i * (1 + sizeof(ItemPointerData)), 1 + (sizeof(ItemPointerData))) == 0))
		{
			leftPos++;
			if (leftPos == numLeftTuples)
				return true;
			leftRelOid = getTableIdOfRelsOidInTSet(fctx, bLeft, leftPos);
		}
	}
	return false;
}

/* Defore a tuple attribute by attribute in each call thus, you do not need to
 * deform the who tuple set or to call SPI_getbin... which starts from the begining
 * and thus, save a lot of processing work! */
bool
heap_deformtuple_iterative(HeapTuple tuple,
						   TupleDesc tupleDesc,
						   Datum *values,
						   char *nulls, deformTupleIterativeState * ds,
						   int tillAttNum, bool finishAll)
{
	if (ds->firstIteration)
	{
		ds->firstIteration = false;
		ds->tup = tuple->t_data;
		ds->att = tupleDesc->attrs;
		ds->tdesc_natts = tupleDesc->natts;
		/* ptr to null bitmask in tuple */
		ds->bp = ds->tup->t_bits;
		/* can we use/set attcacheoff? */
		ds->slow = false;
		ds->natts = HeapTupleHeaderGetNatts(ds->tup);

		/*
		 * In inheritance situations, it is possible that the given tuple
		 * tually has more fields than the caller is expecting.  Don't run the
		 * end of the caller's arrays.
		 */
		ds->natts = Min(ds->natts, ds->tdesc_natts);
		ds->tp = (char *) ds->tup + ds->tup->t_hoff;
		ds->off = 0;
		ds->attnum = 0;
	}
	for (; ds->attnum < ds->natts; ds->attnum++)
	{
		if (HeapTupleHasNulls(tuple) && att_isnull(ds->attnum, ds->bp))
		{
			values[ds->attnum] = (Datum) 0;
			nulls[ds->attnum] = 'n';
			/* can't use attcacheoff anymore */
			ds->slow = true;
			if (!finishAll && ds->attnum >= tillAttNum)
			{
				ds->attnum++;
				if (ds->attnum < ds->natts)
					return true;
				else
					return false;
			}
			else
				continue;
		}
		nulls[ds->attnum] = ' ';
		if (!ds->slow && ds->att[ds->attnum]->attcacheoff >= 0)
			ds->off = ds->att[ds->attnum]->attcacheoff;
		else
		{
			ds->off = att_align(ds->off, ds->att[ds->attnum]->attalign);
			if (!ds->slow)
				ds->att[ds->attnum]->attcacheoff = ds->off;
		}
		values[ds->attnum] = fetchatt(ds->att[ds->attnum], ds->tp + ds->off);
		ds->off = att_addlength(ds->off, ds->att[ds->attnum]->attlen, ds->tp + ds->off);
		if (ds->att[ds->attnum]->attlen <= 0)
			/* can't use attcacheoff anymore */
			ds->slow = true;
		if (!finishAll && ds->attnum >= tillAttNum)
		{
			ds->attnum++;
			if (ds->attnum < ds->natts)
				return true;
			else
				return false;
		}
	}

	/*
	 * If tuple doesn't have all the atts indicated by tupleDesc, read the est
	 * as null
	 */
	for (; ds->attnum < ds->tdesc_natts; ds->attnum++)
	{
		values[ds->attnum] = (Datum) 0;
		nulls[ds->attnum] = 'n';
		if (!finishAll && ds->attnum >= tillAttNum)
		{
			ds->attnum++;
			if (ds->attnum < ds->tdesc_natts)
				return true;
			else
				return false;
		}
	}
	return false;
}

/*
 * Check if the addition of a tuple to a tuple set (TSet) will return a join consistent
 * and connected tuple set(JCC). If it is JCC, then return the result of the addition.
 * tset is the tuple set. tuple is a deformed tuple. tupleDesc is the tuple description.
 * tupleRelID is the relative ID of the tuple's relation in relsOID.
 * dsTuple is a persistant state between calls, that allows to deform the tuple attribute
 * by attribute without restarting the deforming process.
 * Will return null on failure.
 */
TSet
JCCTSetAndTupleAndReturnTSet(alg_fctx * fctx, TSet tset, DTuple tuple
							 ,TupleDesc tupleTupleDesc, int tupleRelID, deformTupleIterativeState * dsTuple)
{
	bytea	   *b = tset->tids;
	int			numAtts = fctx->relsNAtt[tupleRelID];
	int			i = 0;
	int			curTSetAtt = -1;

	for (i = 0; i < numAtts; i++)
	{
		curTSetAtt = fctx->relsAttNumsAtTupleSet[tupleRelID][i];
		if ((tset->n[curTSetAtt] == 'n') || (heap_attisnull(tuple->t, i + 1)))
		{
			if (isResultTupleAttOccursInSomeRelID(fctx, curTSetAtt, b))
			{
				return NULL;
			}
		}
		else
		{
			if (!tuple->deformed)
				heap_deformtuple_iterative(tuple->t, tupleTupleDesc, tuple->v, tuple->n, dsTuple, i, false);
			if (!areAttributesEqual_RT(fctx, curTSetAtt
									   ,tset->v[curTSetAtt]
									   ,tuple->v[i]
									   ))
			{
				return NULL;
			}
		}
	}
	if (!tuple->deformed)
	{
		heap_deformtuple_iterative(tuple->t, tupleTupleDesc, tuple->v, tuple->n, dsTuple, -1, true);
		tuple->tid = (ItemPointer) tuple->v[numAtts];
		tuple->deformed = true;
	}
	for (i = 0; i < numAtts; i++)
	{
		curTSetAtt = fctx->relsAttNumsAtTupleSet[tupleRelID][i];
		if ((tset->n[curTSetAtt] == 'n') && (tuple->n[i] == ' '))
		{
			tset->n[curTSetAtt] = ' ';
			setTSetValue(fctx, tset, tuple->v[i], curTSetAtt);
		}
	}
	bytea	   *outb =
	(bytea *) addRowIDtoTSet_(fctx, b, tuple->tid, tupleRelID);

	setTSetValueNoCopy(fctx, tset, (Datum) outb, LABELS);
	tset->tids = (bytea *) tset->v[LABELS];
	return tset;
}

/*
 * unifies two tuple sets, leftTSet and rightTSet, the vals of the leftTSet are 
 * conviniently in leftTSetVals and nulls in leftTSetNulls. we are
 * also supplied with the information of relationsToCheck which are the restricted
 * relations that we need to check for the JCC (albit limited check).
 * Finalyl the relations of the left tuple set and the right tuple set are also supplied
 * in a bit array.
 * returns null on failure.
 * */
TSet
unifyAndJCC(alg_fctx * fctx, TSet leftTSet, Datum *leftTSetVals, char *leftTSetNulls,
			TSet rightTSet, RBITS relationsToCheck, RBITS relationsOfLeft, RBITS relationsOfRight)
{
	int			i = 0;
	deformTupleIterativeState dsLeftTSet;

	dsLeftTSet.firstIteration = true;
	bool		isLeftAttNull,
				isRightAttNull;

	for (i = FIRSTATT; i < NUMATT; i++)
	{
		isLeftAttNull = (leftTSetNulls[i] == 'n');
		isRightAttNull = (rightTSet->n[i] == 'n');
		if (isLeftAttNull && isRightAttNull)
		{
			if (isResultTupleAttOccursInSomeRelIDWithSpecifiedRelations(fctx, i,
														   relationsToCheck))
			{
				return NULL;
			}
		}
		else if (isLeftAttNull && (!isRightAttNull))
		{
			if (isResultTupleAttOccursInSomeRelIDWithSpecifiedRelations(fctx, i,
									   (relationsToCheck & relationsOfLeft)))
			{
				return NULL;
			}
		}
		else if ((!isLeftAttNull) && isRightAttNull)
		{
			if (isResultTupleAttOccursInSomeRelIDWithSpecifiedRelations(fctx, i,
									  (relationsToCheck & relationsOfRight)))
			{
				return NULL;
			}
		}
		else
		{
			if (!areAttributesEqual_RT(fctx, i, leftTSetVals[i], rightTSet->v[i]))
			{
				return NULL;
			}
		}
	}
	for (i = FIRSTATT; i < NUMATT; i++)
	{
		if ((leftTSetNulls[i] == 'n') && (rightTSet->n[i] == ' '))
		{
			leftTSetNulls[i] = ' ';
			setTSetValue(fctx, leftTSet, rightTSet->v[i], i);
		}
	}
	bytea	   *outb = mergeRowIDs(fctx, leftTSet->tids, rightTSet->tids);

	setTSetValueNoCopy(fctx, leftTSet, (Datum) outb, LABELS);
	leftTSet->tids = (bytea *) outb;
	return leftTSet;
}

/* simple unify, unified two tuple sets without checking for jcc since we assume
 * here that the result will be JCC */
TSet
simpleUnify(alg_fctx * fctx, TSet leftTSet, Datum *leftTSetVals,
			char *leftTSetNulls, TSet rightTSet)
{
	int			i = 0;

	for (i = FIRSTATT; i < NUMATT; i++)
	{
		if ((leftTSetNulls[i] == 'n') && (rightTSet->n[i] == ' '))
		{
			leftTSetNulls[i] = ' ';
			setTSetValue(fctx, leftTSet, rightTSet->v[i], i);
		}
	}
	bytea	   *outb = mergeRowIDs(fctx, leftTSet->tids, rightTSet->tids);

	setTSetValueNoCopy(fctx, leftTSet, (Datum) outb, LABELS);
	leftTSet->tids = (bytea *) outb;
	return leftTSet;
}


/* here we create a singleton tuple set that contains naturally a single tuple.
 * The tuple is a deformed tuple called inBaseTuple. */
TSet
create_tuple_set(alg_fctx * fctx, DTuple inBaseTuple
				 ,TupleDesc baseTupleDesc, int baseTupleInRelsOid)
{
	int			i;
	TSet		returnTSet = newTSet_(fctx);

	returnTSet->n[LABELS] = ' ';
	for (i = 0; i < fctx->relsNAtt[baseTupleInRelsOid]; i++)
	{
		if (inBaseTuple->n[i] == ' ')
		{
			setTSetValue(fctx, returnTSet, inBaseTuple->v[i], fctx->relsAttNumsAtTupleSet[baseTupleInRelsOid][i]);
			returnTSet->n[fctx->relsAttNumsAtTupleSet[baseTupleInRelsOid][i]] = ' ';
		}
	}
	bytea	   *b = createNewTSetRowIDs(fctx, inBaseTuple->tid, baseTupleInRelsOid);

	setTSetValueNoCopy(fctx, returnTSet, (Datum) b, LABELS);
	returnTSet->tids = b;
	return returnTSet;
}

/* check for jcc between two tuples. one tuple is supplied as dtuple and the other tuple
 * is contained in a tuple set tset and pointed by tupleRelIDInTSet in the tset, i.e., it is
 * tupleRelIDInTSet tuples from the start of the tset.*/
bool
JCCTupleInTSetAndDTuple(alg_fctx * fctx, DTuple dtuple, int dtupleRelID,
						TSet tset, int tupleRelIDInTSet)
{
	int			dtupleNumAtts = fctx->relsNAtt[dtupleRelID];
	int			tsetTupleNumAtts = fctx->relsNAtt[tupleRelIDInTSet];
	int		  **dtupleAtts = fctx->sortedRelsAttNumsAtTupleSet[dtupleRelID];
	int		  **tsetTupleAtts = fctx->sortedRelsAttNumsAtTupleSet[tupleRelIDInTSet];
	int			i = 0,
				j = 0;

	while (true)
	{
		if (dtupleAtts[i][0] == tsetTupleAtts[j][0])
		{
			if ((tset->n[tsetTupleAtts[j][0]] == 'n') || (dtuple->n[dtupleAtts[i][1]] == 'n'))
				return false;
			if (!areAttributesEqual_RT(fctx, tsetTupleAtts[j][0], tset->v[tsetTupleAtts[j][0]]
									   ,dtuple->v[dtupleAtts[i][1]]))
			{
				return false;
			}
			i++;
			j++;
			if ((i == dtupleNumAtts) || (j == tsetTupleNumAtts))
				return true;
		}
		else if (dtupleAtts[i][0] < tsetTupleAtts[j][0])
		{
			i++;
			if (i == dtupleNumAtts)
				return true;
		}
		else
		{
			j++;
			if (j == tsetTupleNumAtts)
				return true;
		}
	}
}


/* generating an alternative tuple set is an algorithm terminology which you
 * will need to read in the docs. To generate an alternative tuple set you need
 * a tuple set and a tuple. the tuple set is tset and its relations bit array
 * relsOfTuples. the tuple is the deformed tuple dtuple and its description
 * tupleDesc. tupleRelID is its relative id in relsOID array of relations OIDs. 
 * Basically the idea to take a tuple and create around it a connected component
 * of tuples that we add from the tuple set while keeping the result JCC.
 * */
TSet
generateAlternativeTupleSet(alg_fctx * fctx, TSet tset, RBITS relsOfTuples,
						  DTuple dtuple, TupleDesc tupleDesc, int tupleRelID)
{
  
	/* In phase 1 we find all the JCC tuples in dtset with dtuple */
	RBITS		toCheckForJCC = 0;
	RBITS		tmpBits = relsOfTuples;
	int			i = -1,
				j = -1;

	while (tmpBits)
	{
		i++;
		while (tmpBits && !(varNthBitFromLeft(tmpBits, 0)))
		{
			i++;
			tmpBits <<= 1;
		}
		if (!tmpBits)
			break;
		else
			tmpBits <<= 1;
		j++;
		if (tupleRelID == i)
		{
			makeSureDeformedTuple(fctx, dtuple, tupleDesc, tupleRelID);
			if (!(memcmp(dtuple->tid, TSET_GET_CTID(tset, j), (CTID_SIZE))))
			{
				return NULL;
			}
		}
	}
	toCheckForJCC = fctx->bits_scheme_graph[tupleRelID] & relsOfTuples;
	if (toCheckForJCC == 0)
		return NULL;

	/*
	 * in phase 2 we add all the tuples that are connected to the tuples that
	 * were found JCC without the rejected ones in the component. i.e. we
	 * create a tuple set that is a connected component that contains dtuple.
	 */
	RBITS		tuplesToAddToNewTSet = 0;
	RBITS		tuplesRejected = 0;

	makeSureDeformedTSet(fctx, tset);
	makeSureDeformedTuple(fctx, dtuple, tupleDesc, tupleRelID);
	tmpBits = toCheckForJCC;
	i = -1;
	while (tmpBits)
	{
		i++;
		while (tmpBits && !(varNthBitFromLeft(tmpBits, 0)))
		{
			i++;
			tmpBits <<= 1;
		}
		if (!tmpBits)
			break;
		else
			tmpBits <<= 1;
		if (JCCTupleInTSetAndDTuple(fctx, dtuple, tupleRelID, tset, i))
			varSetNthBitFromLeft(tuplesToAddToNewTSet, i);
		else
			varSetNthBitFromLeft(tuplesRejected, i);
	}
	if (tuplesToAddToNewTSet == 0)
		return NULL;
	varUnSetNthBitFromLeft(tuplesToAddToNewTSet, tupleRelID);
    
	/* find all the rest of the component */
	tmpBits = ~tuplesToAddToNewTSet;
	while (tmpBits != tuplesToAddToNewTSet)
	{
		tmpBits = tuplesToAddToNewTSet;
		for (i = 0; i < fctx->numRels; i++)
			if ((i != tupleRelID) && (!varNthBitFromLeft(tuplesRejected, i))
				&& (varNthBitFromLeft(relsOfTuples, i)))
				if (tuplesToAddToNewTSet & fctx->bits_scheme_graph[i])
					varSetNthBitFromLeft(tuplesToAddToNewTSet, i);
	}
    
	/* Lets create a new tuple set with the base tuple */
	int			curTSetAtt;
	int			numAtts = fctx->relsNAtt[tupleRelID];
	TSet		newTSet = newTSet_(fctx);

	for (i = FIRSTATT; i < NUMATT; i++)
		newTSet->n[i] = 'n';
	newTSet->n[LABELS] = ' ';
	for (i = 0; i < numAtts; i++)
	{
		curTSetAtt = fctx->relsAttNumsAtTupleSet[tupleRelID][i];
		if (dtuple->n[i] == ' ')
		{
			setTSetValue(fctx, newTSet, dtuple->v[i], curTSetAtt);
			newTSet->n[curTSetAtt] = ' ';
		}
	}
    
	/* Lets add the rest of the tuples. */
	int			k;
	int			numTupleInNewTSet = 1;

	varUnSetNthBitFromLeft(tuplesToAddToNewTSet, tupleRelID);
	tmpBits = tuplesToAddToNewTSet;
	i = -1;
	while (tmpBits)
	{
		i++;
		while (tmpBits && !(varNthBitFromLeft(tmpBits, 0)))
		{
			i++;
			tmpBits <<= 1;
		}
		if (!tmpBits)
			break;
		else
			tmpBits <<= 1;
		numTupleInNewTSet++;
		numAtts = fctx->relsNAtt[i];
		for (k = 0; k < numAtts; k++)
		{
			curTSetAtt = fctx->relsAttNumsAtTupleSet[i][k];
			if ((newTSet->n[curTSetAtt] == 'n') && (tset->n[curTSetAtt] == ' '))
			{
				newTSet->n[curTSetAtt] = ' ';
				setTSetValue(fctx, newTSet, tset->v[curTSetAtt], curTSetAtt);
			}
		}
	}
    
	/* finally lets build the tableID:CTID pairs array. */
	int			z = -1;
	RBITS		tmpRelsOfTuples = relsOfTuples;

	tmpBits = tuplesToAddToNewTSet;
	varSetNthBitFromLeft(tmpBits, tupleRelID);
	i = -1, j = -1;
	bytea	   *outb = palloc(VARHDRSZ + (numTupleInNewTSet * TABLEID_CTID_SIZE));

	SET_VARSIZE(outb, VARHDRSZ + (numTupleInNewTSet * TABLEID_CTID_SIZE));
	while (tmpBits)
	{
		i++;
		if (varNthBitFromLeft(tmpRelsOfTuples, 0))
			z++;
		while (tmpBits && !(varNthBitFromLeft(tmpBits, 0)))
		{
			i++;
			tmpBits <<= 1;
			tmpRelsOfTuples <<= 1;
			if (varNthBitFromLeft(tmpRelsOfTuples, 0))
				z++;
		}
		if (!tmpBits)
			break;
		else
		{
			tmpBits <<= 1;
			tmpRelsOfTuples <<= 1;
		}
		j++;
		VARDATA(outb)[j * TABLEID_CTID_SIZE] = (char) i;
		if (i == tupleRelID)
			memcpy(VARDATA(outb) + (j * TABLEID_CTID_SIZE) + 1
				   ,dtuple->tid
				   ,CTID_SIZE);
		else
			memcpy(VARDATA(outb) + (j * TABLEID_CTID_SIZE) + 1
				   ,VARDATA(tset->tids) + (z * TABLEID_CTID_SIZE) + 1
				   ,CTID_SIZE);
	}
	newTSet->v[LABELS] = (Datum) outb;
	newTSet->tids = outb;
	return newTSet;
}

/* Sometimes before we want to store a tuple set we wish to clear its values
 * and leave only the LABELS attribute.*/
HeapTuple
clearValuesFromTSet(alg_fctx * fctx, TSet tset, char *nulls)
{
	Datum		valuesTSet[1];

	valuesTSet[0] = (Datum) tset->tids;
	return (heap_formtuple(fctx->tupleSetDesc, valuesTSet, nulls));
}
